"""
Connection module for Amazon EC2

.. versionadded:: 2015.8.0

:configuration: This module accepts explicit EC2 credentials but can also
    utilize IAM roles assigned to the instance through Instance Profiles.
    Dynamic credentials are then automatically obtained from AWS API and no
    further configuration is necessary. More Information available here__.

.. __: http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html

If IAM roles are not used you need to specify them either in a pillar or
in the minion's config file:

.. code-block:: yaml

    ec2.keyid: GKTADJGHEIQSXMKKRBJ08H
    ec2.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

A region may also be specified in the configuration:

.. code-block:: yaml

    ec2.region: us-east-1

If a region is not specified, the default is us-east-1.

It's also possible to specify key, keyid, and region via a profile, either
as a passed in dict, or as a string to pull from pillars or minion config:

.. code-block:: yaml

    myprofile:
      keyid: GKTADJGHEIQSXMKKRBJ08H
      key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
      region: us-east-1

:depends: boto

"""
# keep lint from choking on _get_conn and _cache_id
# pylint: disable=E0602


import logging
import time

import salt.utils.compat
import salt.utils.data
import salt.utils.json
import salt.utils.versions
from salt.exceptions import CommandExecutionError, SaltInvocationError

try:
    # pylint: disable=unused-import
    import boto
    import boto.ec2

    # pylint: enable=unused-import
    from boto.ec2.blockdevicemapping import BlockDeviceMapping, BlockDeviceType
    from boto.ec2.networkinterface import (
        NetworkInterfaceSpecification,
        NetworkInterfaceCollection,
    )

    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False


log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if boto libraries exist and if boto libraries are greater than
    a given version.
    """
    # the boto_ec2 execution module relies on the connect_to_region() method
    # which was added in boto 2.8.0
    # https://github.com/boto/boto/commit/33ac26b416fbb48a60602542b4ce15dcc7029f12
    has_boto_reqs = salt.utils.versions.check_boto_reqs(
        boto_ver="2.8.0", check_boto3=False
    )
    if has_boto_reqs is True:
        __utils__["boto.assign_funcs"](__name__, "ec2", pack=__salt__)
    return has_boto_reqs


def __init__(opts):
    if HAS_BOTO:
        __utils__["boto.assign_funcs"](__name__, "ec2")


def _get_all_eip_addresses(
    addresses=None, allocation_ids=None, region=None, key=None, keyid=None, profile=None
):
    """
    Get all EIP's associated with the current credentials.

    addresses
        (list) - Optional list of addresses.  If provided, only those those in the
        list will be returned.
    allocation_ids
        (list) - Optional list of allocation IDs.  If provided, only the
        addresses associated with the given allocation IDs will be returned.

    returns
        (list) - The requested Addresses as a list of :class:`boto.ec2.address.Address`
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    try:
        return conn.get_all_addresses(
            addresses=addresses, allocation_ids=allocation_ids
        )
    except boto.exception.BotoServerError as e:
        log.error(e)
        return []


def get_all_eip_addresses(
    addresses=None, allocation_ids=None, region=None, key=None, keyid=None, profile=None
):
    """
    Get public addresses of some, or all EIPs associated with the current account.

    addresses
        (list) - Optional list of addresses.  If provided, only the addresses
        associated with those in the list will be returned.
    allocation_ids
        (list) - Optional list of allocation IDs.  If provided, only the
        addresses associated with the given allocation IDs will be returned.

    returns
        (list) - A list of the requested EIP addresses

    CLI Example:

    .. code-block:: bash

        salt-call boto_ec2.get_all_eip_addresses

    .. versionadded:: 2016.3.0
    """
    return [
        x.public_ip
        for x in _get_all_eip_addresses(
            addresses, allocation_ids, region, key, keyid, profile
        )
    ]


def get_unassociated_eip_address(
    domain="standard", region=None, key=None, keyid=None, profile=None
):
    """
    Return the first unassociated EIP

    domain
        Indicates whether the address is an EC2 address or a VPC address
        (standard|vpc).

    CLI Example:

    .. code-block:: bash

        salt-call boto_ec2.get_unassociated_eip_address

    .. versionadded:: 2016.3.0
    """
    eip = None
    for address in get_all_eip_addresses(
        region=region, key=key, keyid=keyid, profile=profile
    ):
        address_info = get_eip_address_info(
            addresses=address, region=region, key=key, keyid=keyid, profile=profile
        )[0]
        if address_info["instance_id"]:
            log.debug(
                "%s is already associated with the instance %s",
                address,
                address_info["instance_id"],
            )
            continue

        if address_info["network_interface_id"]:
            log.debug(
                "%s is already associated with the network interface %s",
                address,
                address_info["network_interface_id"],
            )
            continue

        if address_info["domain"] == domain:
            log.debug(
                "The first unassociated EIP address in the domain '%s' is %s",
                domain,
                address,
            )
            eip = address
            break

    if not eip:
        log.debug("No unassociated Elastic IP found!")

    return eip


def get_eip_address_info(
    addresses=None, allocation_ids=None, region=None, key=None, keyid=None, profile=None
):
    """
    Get 'interesting' info about some, or all EIPs associated with the current account.

    addresses
        (list) - Optional list of addresses.  If provided, only the addresses
        associated with those in the list will be returned.
    allocation_ids
        (list) - Optional list of allocation IDs.  If provided, only the
        addresses associated with the given allocation IDs will be returned.

    returns
        (list of dicts) - A list of dicts, each containing the info for one of the requested EIPs.

    CLI Example:

    .. code-block:: bash

        salt-call boto_ec2.get_eip_address_info addresses=52.4.2.15

    .. versionadded:: 2016.3.0
    """
    if type(addresses) == (type("string")):
        addresses = [addresses]
    if type(allocation_ids) == (type("string")):
        allocation_ids = [allocation_ids]

    ret = _get_all_eip_addresses(
        addresses=addresses,
        allocation_ids=allocation_ids,
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
    )

    interesting = [
        "allocation_id",
        "association_id",
        "domain",
        "instance_id",
        "network_interface_id",
        "network_interface_owner_id",
        "public_ip",
        "private_ip_address",
    ]

    return [{x: getattr(address, x) for x in interesting} for address in ret]


def allocate_eip_address(domain=None, region=None, key=None, keyid=None, profile=None):
    """
    Allocate a new Elastic IP address and associate it with your account.

    domain
        (string) Optional param - if set to exactly 'vpc', the address will be
        allocated to the VPC.  The default simply maps the EIP to your
        account container.

    returns
        (dict) dict of 'interesting' information about the newly allocated EIP,
        with probably the most interesting keys being 'public_ip'; and
        'allocation_id' iff 'domain=vpc' was passed.

    CLI Example:

    .. code-block:: bash

        salt-call boto_ec2.allocate_eip_address domain=vpc

    .. versionadded:: 2016.3.0
    """
    if domain and domain != "vpc":
        raise SaltInvocationError(
            "The only permitted value for the 'domain' param is 'vpc'."
        )

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    try:
        address = conn.allocate_address(domain=domain)
    except boto.exception.BotoServerError as e:
        log.error(e)
        return False

    interesting = [
        "allocation_id",
        "association_id",
        "domain",
        "instance_id",
        "network_interface_id",
        "network_interface_owner_id",
        "public_ip",
        "private_ip_address",
    ]

    return {x: getattr(address, x) for x in interesting}


def release_eip_address(
    public_ip=None, allocation_id=None, region=None, key=None, keyid=None, profile=None
):
    """
    Free an Elastic IP address.  Pass either a public IP address to release an
    EC2 Classic EIP, or an AllocationId to release a VPC EIP.

    public_ip
        (string) - The public IP address - for EC2 elastic IPs.
    allocation_id
        (string) - The Allocation ID - for VPC elastic IPs.

    returns
        (bool) - True on success, False on failure

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.release_eip_address allocation_id=eipalloc-ef382c8a

    .. versionadded:: 2016.3.0
    """
    if not salt.utils.data.exactly_one((public_ip, allocation_id)):
        raise SaltInvocationError(
            "Exactly one of 'public_ip' OR 'allocation_id' must be provided"
        )

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    try:
        return conn.release_address(public_ip, allocation_id)
    except boto.exception.BotoServerError as e:
        log.error(e)
        return False


def associate_eip_address(
    instance_id=None,
    instance_name=None,
    public_ip=None,
    allocation_id=None,
    network_interface_id=None,
    network_interface_name=None,
    private_ip_address=None,
    allow_reassociation=False,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Associate an Elastic IP address with a currently running instance or a network interface.
    This requires exactly one of either 'public_ip' or 'allocation_id', depending
    on whether you’re associating a VPC address or a plain EC2 address.

    instance_id
        (string) – ID of the instance to associate with (exclusive with 'instance_name')
    instance_name
        (string) – Name tag of the instance to associate with (exclusive with 'instance_id')
    public_ip
        (string) – Public IP address, for standard EC2 based allocations.
    allocation_id
        (string) – Allocation ID for a VPC-based EIP.
    network_interface_id
        (string) - ID of the network interface to associate the EIP with
    network_interface_name
        (string) - Name of the network interface to associate the EIP with
    private_ip_address
        (string) – The primary or secondary private IP address to associate with the Elastic IP address.
    allow_reassociation
        (bool)   – Allow a currently associated EIP to be re-associated with the new instance or interface.

    returns
        (bool)   - True on success, False on failure.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.associate_eip_address instance_name=bubba.ho.tep allocation_id=eipalloc-ef382c8a

    .. versionadded:: 2016.3.0
    """
    if not salt.utils.data.exactly_one(
        (instance_id, instance_name, network_interface_id, network_interface_name)
    ):
        raise SaltInvocationError(
            "Exactly one of 'instance_id', "
            "'instance_name', 'network_interface_id', "
            "'network_interface_name' must be provided"
        )

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    if instance_name:
        try:
            instance_id = get_id(
                name=instance_name, region=region, key=key, keyid=keyid, profile=profile
            )
        except boto.exception.BotoServerError as e:
            log.error(e)
            return False
        if not instance_id:
            log.error(
                "Given instance_name '%s' cannot be mapped to an instance_id",
                instance_name,
            )
            return False

    if network_interface_name:
        try:
            network_interface_id = get_network_interface_id(
                network_interface_name,
                region=region,
                key=key,
                keyid=keyid,
                profile=profile,
            )
        except boto.exception.BotoServerError as e:
            log.error(e)
            return False
        if not network_interface_id:
            log.error(
                "Given network_interface_name '%s' cannot be mapped to "
                "an network_interface_id",
                network_interface_name,
            )
            return False

    try:
        return conn.associate_address(
            instance_id=instance_id,
            public_ip=public_ip,
            allocation_id=allocation_id,
            network_interface_id=network_interface_id,
            private_ip_address=private_ip_address,
            allow_reassociation=allow_reassociation,
        )
    except boto.exception.BotoServerError as e:
        log.error(e)
        return False


def disassociate_eip_address(
    public_ip=None, association_id=None, region=None, key=None, keyid=None, profile=None
):
    """
    Disassociate an Elastic IP address from a currently running instance. This
    requires exactly one of either 'association_id' or 'public_ip', depending
    on whether you’re dealing with a VPC or EC2 Classic address.

    public_ip
        (string) – Public IP address, for EC2 Classic allocations.
    association_id
        (string) – Association ID for a VPC-bound EIP.

    returns
        (bool)   - True on success, False on failure.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.disassociate_eip_address association_id=eipassoc-e3ba2d16

    .. versionadded:: 2016.3.0
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    try:
        return conn.disassociate_address(public_ip, association_id)
    except boto.exception.BotoServerError as e:
        log.error(e)
        return False


def assign_private_ip_addresses(
    network_interface_name=None,
    network_interface_id=None,
    private_ip_addresses=None,
    secondary_private_ip_address_count=None,
    allow_reassignment=False,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Assigns one or more secondary private IP addresses to a network interface.

    network_interface_id
        (string) - ID of the network interface to associate the IP with (exclusive with 'network_interface_name')
    network_interface_name
        (string) - Name of the network interface to associate the IP with (exclusive with 'network_interface_id')
    private_ip_addresses
        (list) - Assigns the specified IP addresses as secondary IP addresses to the network interface (exclusive with 'secondary_private_ip_address_count')
    secondary_private_ip_address_count
        (int) - The number of secondary IP addresses to assign to the network interface. (exclusive with 'private_ip_addresses')
    allow_reassociation
        (bool)   – Allow a currently associated EIP to be re-associated with the new instance or interface.

    returns
        (bool)   - True on success, False on failure.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.assign_private_ip_addresses network_interface_name=my_eni private_ip_addresses=private_ip
        salt myminion boto_ec2.assign_private_ip_addresses network_interface_name=my_eni secondary_private_ip_address_count=2

    .. versionadded:: 2017.7.0
    """
    if not salt.utils.data.exactly_one((network_interface_name, network_interface_id)):
        raise SaltInvocationError(
            "Exactly one of 'network_interface_name', "
            "'network_interface_id' must be provided"
        )

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    if network_interface_name:
        try:
            network_interface_id = get_network_interface_id(
                network_interface_name,
                region=region,
                key=key,
                keyid=keyid,
                profile=profile,
            )
        except boto.exception.BotoServerError as e:
            log.error(e)
            return False
        if not network_interface_id:
            log.error(
                "Given network_interface_name '%s' cannot be mapped to "
                "an network_interface_id",
                network_interface_name,
            )
            return False

    try:
        return conn.assign_private_ip_addresses(
            network_interface_id=network_interface_id,
            private_ip_addresses=private_ip_addresses,
            secondary_private_ip_address_count=secondary_private_ip_address_count,
            allow_reassignment=allow_reassignment,
        )
    except boto.exception.BotoServerError as e:
        log.error(e)
        return False


def unassign_private_ip_addresses(
    network_interface_name=None,
    network_interface_id=None,
    private_ip_addresses=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Unassigns one or more secondary private IP addresses from a network interface

    network_interface_id
        (string) - ID of the network interface to associate the IP with (exclusive with 'network_interface_name')
    network_interface_name
        (string) - Name of the network interface to associate the IP with (exclusive with 'network_interface_id')
    private_ip_addresses
        (list) - Assigns the specified IP addresses as secondary IP addresses to the network interface.

    returns
        (bool)   - True on success, False on failure.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.unassign_private_ip_addresses network_interface_name=my_eni private_ip_addresses=private_ip

    .. versionadded:: 2017.7.0
    """
    if not salt.utils.data.exactly_one((network_interface_name, network_interface_id)):
        raise SaltInvocationError(
            "Exactly one of 'network_interface_name', "
            "'network_interface_id' must be provided"
        )

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    if network_interface_name:
        try:
            network_interface_id = get_network_interface_id(
                network_interface_name,
                region=region,
                key=key,
                keyid=keyid,
                profile=profile,
            )
        except boto.exception.BotoServerError as e:
            log.error(e)
            return False
        if not network_interface_id:
            log.error(
                "Given network_interface_name '%s' cannot be mapped to "
                "an network_interface_id",
                network_interface_name,
            )
            return False

    try:
        return conn.unassign_private_ip_addresses(
            network_interface_id=network_interface_id,
            private_ip_addresses=private_ip_addresses,
        )
    except boto.exception.BotoServerError as e:
        log.error(e)
        return False


def get_zones(region=None, key=None, keyid=None, profile=None):
    """
    Get a list of AZs for the configured region.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.get_zones
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    return [z.name for z in conn.get_all_zones()]


def find_instances(
    instance_id=None,
    name=None,
    tags=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
    return_objs=False,
    in_states=None,
    filters=None,
):

    """
    Given instance properties, find and return matching instance ids

    CLI Examples:

    .. code-block:: bash

        salt myminion boto_ec2.find_instances # Lists all instances
        salt myminion boto_ec2.find_instances name=myinstance
        salt myminion boto_ec2.find_instances tags='{"mytag": "value"}'
        salt myminion boto_ec2.find_instances filters='{"vpc-id": "vpc-12345678"}'

    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    try:
        filter_parameters = {"filters": {}}

        if instance_id:
            filter_parameters["instance_ids"] = [instance_id]

        if name:
            filter_parameters["filters"]["tag:Name"] = name

        if tags:
            for tag_name, tag_value in tags.items():
                filter_parameters["filters"]["tag:{}".format(tag_name)] = tag_value

        if filters:
            filter_parameters["filters"].update(filters)

        reservations = conn.get_all_reservations(**filter_parameters)
        instances = [i for r in reservations for i in r.instances]
        log.debug(
            "The filters criteria %s matched the following instances:%s",
            filter_parameters,
            instances,
        )

        if in_states:
            instances = [i for i in instances if i.state in in_states]
            log.debug(
                "Limiting instance matches to those in the requested states: %s",
                instances,
            )
        if instances:
            if return_objs:
                return instances
            return [instance.id for instance in instances]
        else:
            return []
    except boto.exception.BotoServerError as exc:
        log.error(exc)
        return []


def create_image(
    ami_name,
    instance_id=None,
    instance_name=None,
    tags=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
    description=None,
    no_reboot=False,
    dry_run=False,
    filters=None,
):
    """
    Given instance properties that define exactly one instance, create AMI and return AMI-id.

    CLI Examples:

    .. code-block:: bash

        salt myminion boto_ec2.create_image ami_name instance_name=myinstance
        salt myminion boto_ec2.create_image another_ami_name tags='{"mytag": "value"}' description='this is my ami'

    """

    instances = find_instances(
        instance_id=instance_id,
        name=instance_name,
        tags=tags,
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
        return_objs=True,
        filters=filters,
    )

    if not instances:
        log.error("Source instance not found")
        return False
    if len(instances) > 1:
        log.error(
            "Multiple instances found, must match exactly only one instance to create"
            " an image from"
        )
        return False

    instance = instances[0]
    try:
        return instance.create_image(
            ami_name, description=description, no_reboot=no_reboot, dry_run=dry_run
        )
    except boto.exception.BotoServerError as exc:
        log.error(exc)
        return False


def find_images(
    ami_name=None,
    executable_by=None,
    owners=None,
    image_ids=None,
    tags=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
    return_objs=False,
):

    """
    Given image properties, find and return matching AMI ids

    CLI Examples:

    .. code-block:: bash

        salt myminion boto_ec2.find_images tags='{"mytag": "value"}'

    """
    retries = 30
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    while retries:
        try:
            filter_parameters = {"filters": {}}
            if image_ids:
                filter_parameters["image_ids"] = [image_ids]
            if executable_by:
                filter_parameters["executable_by"] = [executable_by]
            if owners:
                filter_parameters["owners"] = [owners]
            if ami_name:
                filter_parameters["filters"]["name"] = ami_name
            if tags:
                for tag_name, tag_value in tags.items():
                    filter_parameters["filters"]["tag:{}".format(tag_name)] = tag_value
            images = conn.get_all_images(**filter_parameters)
            log.debug(
                "The filters criteria %s matched the following images:%s",
                filter_parameters,
                images,
            )
            if images:
                if return_objs:
                    return images
                return [image.id for image in images]
            else:
                return False
        except boto.exception.BotoServerError as exc:
            if exc.error_code == "Throttling":
                log.debug("Throttled by AWS API, will retry in 5 seconds...")
                time.sleep(5)
                retries -= 1
                continue
            log.error("Failed to convert AMI name `%s` to an AMI ID: %s", ami_name, exc)
            return False
    return False


def terminate(
    instance_id=None,
    name=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
    filters=None,
):
    """
    Terminate the instance described by instance_id or name.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.terminate name=myinstance
        salt myminion boto_ec2.terminate instance_id=i-a46b9f
    """
    instances = find_instances(
        instance_id=instance_id,
        name=name,
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
        return_objs=True,
        filters=filters,
    )
    if instances in (False, None, []):
        return instances

    if len(instances) == 1:
        instances[0].terminate()
        return True
    else:
        log.warning("Refusing to terminate multiple instances at once")
        return False


def get_id(
    name=None,
    tags=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
    in_states=None,
    filters=None,
):

    """
    Given instance properties, return the instance id if it exists.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.get_id myinstance

    """
    instance_ids = find_instances(
        name=name,
        tags=tags,
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
        in_states=in_states,
        filters=filters,
    )
    if instance_ids:
        log.info("Instance ids: %s", " ".join(instance_ids))
        if len(instance_ids) == 1:
            return instance_ids[0]
        else:
            raise CommandExecutionError(
                "Found more than one instance matching the criteria."
            )
    else:
        log.warning("Could not find instance.")
        return None


def get_tags(instance_id=None, keyid=None, key=None, profile=None, region=None):
    """
    Given an instance_id, return a list of tags associated with that instance.

    returns
        (list) - list of tags as key/value pairs

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.get_tags instance_id
    """
    tags = []
    client = _get_conn(key=key, keyid=keyid, profile=profile, region=region)
    result = client.get_all_tags(filters={"resource-id": instance_id})
    if result:
        for tag in result:
            tags.append({tag.name: tag.value})
    else:
        log.info("No tags found for instance_id %s", instance_id)
    return tags


def exists(
    instance_id=None,
    name=None,
    tags=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
    in_states=None,
    filters=None,
):
    """
    Given an instance id, check to see if the given instance id exists.

    Returns True if the given instance with the given id, name, or tags
    exists; otherwise, False is returned.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.exists myinstance
    """
    instances = find_instances(
        instance_id=instance_id,
        name=name,
        tags=tags,
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
        in_states=in_states,
        filters=filters,
    )
    if instances:
        log.info("Instance exists.")
        return True
    else:
        log.warning("Instance does not exist.")
        return False


def _to_blockdev_map(thing):
    """
    Convert a string, or a json payload, or a dict in the right
    format, into a boto.ec2.blockdevicemapping.BlockDeviceMapping as
    needed by instance_present().  The following YAML is a direct
    representation of what is expected by the underlying boto EC2 code.

    YAML example:

    .. code-block:: yaml

        device-maps:
            /dev/sdb:
                ephemeral_name: ephemeral0
            /dev/sdc:
                ephemeral_name: ephemeral1
            /dev/sdd:
                ephemeral_name: ephemeral2
            /dev/sde:
                ephemeral_name: ephemeral3
            /dev/sdf:
                size: 20
                volume_type: gp2

    """
    if not thing:
        return None
    if isinstance(thing, BlockDeviceMapping):
        return thing
    if isinstance(thing, str):
        thing = salt.utils.json.loads(thing)
    if not isinstance(thing, dict):
        log.error(
            "Can't convert '%s' of type %s to a "
            "boto.ec2.blockdevicemapping.BlockDeviceMapping",
            thing,
            type(thing),
        )
        return None

    bdm = BlockDeviceMapping()
    for d, t in thing.items():
        bdt = BlockDeviceType(
            ephemeral_name=t.get("ephemeral_name"),
            no_device=t.get("no_device", False),
            volume_id=t.get("volume_id"),
            snapshot_id=t.get("snapshot_id"),
            status=t.get("status"),
            attach_time=t.get("attach_time"),
            delete_on_termination=t.get("delete_on_termination", False),
            size=t.get("size"),
            volume_type=t.get("volume_type"),
            iops=t.get("iops"),
            encrypted=t.get("encrypted"),
        )
        bdm[d] = bdt

    return bdm


def run(
    image_id,
    name=None,
    tags=None,
    key_name=None,
    security_groups=None,
    user_data=None,
    instance_type="m1.small",
    placement=None,
    kernel_id=None,
    ramdisk_id=None,
    monitoring_enabled=None,
    vpc_id=None,
    vpc_name=None,
    subnet_id=None,
    subnet_name=None,
    private_ip_address=None,
    block_device_map=None,
    disable_api_termination=None,
    instance_initiated_shutdown_behavior=None,
    placement_group=None,
    client_token=None,
    security_group_ids=None,
    security_group_names=None,
    additional_info=None,
    tenancy=None,
    instance_profile_arn=None,
    instance_profile_name=None,
    ebs_optimized=None,
    network_interface_id=None,
    network_interface_name=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
    network_interfaces=None,
):
    # TODO: support multi-instance reservations
    """
    Create and start an EC2 instance.

    Returns True if the instance was created; otherwise False.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.run ami-b80c2b87 name=myinstance

    image_id
        (string) – The ID of the image to run.
    name
        (string) - The name of the instance.
    tags
        (dict of key: value pairs) - tags to apply to the instance.
    key_name
        (string) – The name of the key pair with which to launch instances.
    security_groups
        (list of strings) – The names of the EC2 classic security groups with
        which to associate instances
    user_data
        (string) – The Base64-encoded MIME user data to be made available to the
        instance(s) in this reservation.
    instance_type
        (string) – The type of instance to run.  Note that some image types
        (e.g. hvm) only run on some instance types.
    placement
        (string) – The Availability Zone to launch the instance into.
    kernel_id
        (string) – The ID of the kernel with which to launch the instances.
    ramdisk_id
        (string) – The ID of the RAM disk with which to launch the instances.
    monitoring_enabled
        (bool) – Enable detailed CloudWatch monitoring on the instance.
    vpc_id
        (string) - ID of a VPC to bind the instance to.  Exclusive with vpc_name.
    vpc_name
        (string) - Name of a VPC to bind the instance to.  Exclusive with vpc_id.
    subnet_id
        (string) – The subnet ID within which to launch the instances for VPC.
    subnet_name
        (string) – The name of a subnet within which to launch the instances for VPC.
    private_ip_address
        (string) – If you’re using VPC, you can optionally use this parameter to
        assign the instance a specific available IP address from the subnet
        (e.g. 10.0.0.25).
    block_device_map
        (boto.ec2.blockdevicemapping.BlockDeviceMapping) – A BlockDeviceMapping
        data structure describing the EBS volumes associated with the Image.
        (string) - A string representation of a BlockDeviceMapping structure
        (dict) - A dict describing a BlockDeviceMapping structure

        YAML example:

        .. code-block:: yaml

            device-maps:
                /dev/sdb:
                    ephemeral_name: ephemeral0
                /dev/sdc:
                    ephemeral_name: ephemeral1
                /dev/sdd:
                    ephemeral_name: ephemeral2
                /dev/sde:
                    ephemeral_name: ephemeral3
                /dev/sdf:
                    size: 20
                    volume_type: gp2

    disable_api_termination
        (bool) – If True, the instances will be locked and will not be able to
        be terminated via the API.
    instance_initiated_shutdown_behavior
        (string) – Specifies whether the instance stops or terminates on
        instance-initiated shutdown. Valid values are: stop, terminate
    placement_group
        (string) – If specified, this is the name of the placement group in
        which the instance(s) will be launched.
    client_token
        (string) – Unique, case-sensitive identifier you provide to ensure
        idempotency of the request. Maximum 64 ASCII characters.
    security_group_ids
        (list of strings) – The ID(s) of the VPC security groups with which to
        associate instances.
    security_group_names
        (list of strings) – The name(s) of the VPC security groups with which to
        associate instances.
    additional_info
        (string) – Specifies additional information to make available to the
        instance(s).
    tenancy
        (string) – The tenancy of the instance you want to launch. An instance
        with a tenancy of ‘dedicated’ runs on single-tenant hardware and can
        only be launched into a VPC. Valid values are:”default” or “dedicated”.
        NOTE: To use dedicated tenancy you MUST specify a VPC subnet-ID as well.
    instance_profile_arn
        (string) – The Amazon resource name (ARN) of the IAM Instance Profile
        (IIP) to associate with the instances.
    instance_profile_name
        (string) – The name of the IAM Instance Profile (IIP) to associate with
        the instances.
    ebs_optimized
        (bool) – Whether the instance is optimized for EBS I/O. This
        optimization provides dedicated throughput to Amazon EBS and an
        optimized configuration stack to provide optimal EBS I/O performance.
        This optimization isn’t available with all instance types.
    network_interfaces
        (boto.ec2.networkinterface.NetworkInterfaceCollection) – A
        NetworkInterfaceCollection data structure containing the ENI
        specifications for the instance.
    network_interface_id
        (string) - ID of the network interface to attach to the instance
    network_interface_name
        (string) - Name of the network interface to attach to the instance

    """
    if all((subnet_id, subnet_name)):
        raise SaltInvocationError(
            "Only one of subnet_name or subnet_id may be provided."
        )
    if subnet_name:
        r = __salt__["boto_vpc.get_resource_id"](
            "subnet", subnet_name, region=region, key=key, keyid=keyid, profile=profile
        )
        if "id" not in r:
            log.warning("Couldn't resolve subnet name %s.", subnet_name)
            return False
        subnet_id = r["id"]

    if all((security_group_ids, security_group_names)):
        raise SaltInvocationError(
            "Only one of security_group_ids or security_group_names may be provided."
        )
    if security_group_names:
        security_group_ids = []
        for sgn in security_group_names:
            r = __salt__["boto_secgroup.get_group_id"](
                sgn,
                vpc_name=vpc_name,
                region=region,
                key=key,
                keyid=keyid,
                profile=profile,
            )
            if not r:
                log.warning("Couldn't resolve security group name %s", sgn)
                return False
            security_group_ids += [r]

    network_interface_args = list(
        map(
            int,
            [
                network_interface_id is not None,
                network_interface_name is not None,
                network_interfaces is not None,
            ],
        )
    )

    if sum(network_interface_args) > 1:
        raise SaltInvocationError(
            "Only one of network_interface_id, "
            "network_interface_name or "
            "network_interfaces may be provided."
        )
    if network_interface_name:
        result = get_network_interface_id(
            network_interface_name, region=region, key=key, keyid=keyid, profile=profile
        )
        network_interface_id = result["result"]
        if not network_interface_id:
            log.warning(
                "Given network_interface_name '%s' cannot be mapped to an "
                "network_interface_id",
                network_interface_name,
            )

    if network_interface_id:
        interface = NetworkInterfaceSpecification(
            network_interface_id=network_interface_id, device_index=0
        )
    else:
        interface = NetworkInterfaceSpecification(
            subnet_id=subnet_id, groups=security_group_ids, device_index=0
        )

    if network_interfaces:
        interfaces_specs = [
            NetworkInterfaceSpecification(**x) for x in network_interfaces
        ]
        interfaces = NetworkInterfaceCollection(*interfaces_specs)
    else:
        interfaces = NetworkInterfaceCollection(interface)

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    reservation = conn.run_instances(
        image_id,
        key_name=key_name,
        security_groups=security_groups,
        user_data=user_data,
        instance_type=instance_type,
        placement=placement,
        kernel_id=kernel_id,
        ramdisk_id=ramdisk_id,
        monitoring_enabled=monitoring_enabled,
        private_ip_address=private_ip_address,
        block_device_map=_to_blockdev_map(block_device_map),
        disable_api_termination=disable_api_termination,
        instance_initiated_shutdown_behavior=instance_initiated_shutdown_behavior,
        placement_group=placement_group,
        client_token=client_token,
        additional_info=additional_info,
        tenancy=tenancy,
        instance_profile_arn=instance_profile_arn,
        instance_profile_name=instance_profile_name,
        ebs_optimized=ebs_optimized,
        network_interfaces=interfaces,
    )
    if not reservation:
        log.warning("Instance could not be reserved")
        return False

    instance = reservation.instances[0]

    status = "pending"
    while status == "pending":
        time.sleep(5)
        status = instance.update()
    if status == "running":
        if name:
            instance.add_tag("Name", name)
        if tags:
            instance.add_tags(tags)
        return {"instance_id": instance.id}
    else:
        log.warning('Instance could not be started -- status is "%s"', status)


def get_key(key_name, region=None, key=None, keyid=None, profile=None):
    """
    Check to see if a key exists. Returns fingerprint and name if
    it does and False if it doesn't

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.get_key mykey
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    try:
        key = conn.get_key_pair(key_name)
        log.debug("the key to return is : %s", key)
        if key is None:
            return False
        return key.name, key.fingerprint
    except boto.exception.BotoServerError as e:
        log.debug(e)
        return False


def create_key(key_name, save_path, region=None, key=None, keyid=None, profile=None):
    """
    Creates a key and saves it to a given path.
    Returns the private key.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.create_key mykey /root/
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    try:
        key = conn.create_key_pair(key_name)
        log.debug("the key to return is : %s", key)
        key.save(save_path)
        return key.material
    except boto.exception.BotoServerError as e:
        log.debug(e)
        return False


def import_key(
    key_name, public_key_material, region=None, key=None, keyid=None, profile=None
):
    """
    Imports the public key from an RSA key pair that you created with a third-party tool.
    Supported formats:
    - OpenSSH public key format (e.g., the format in ~/.ssh/authorized_keys)
    - Base64 encoded DER format
    - SSH public key file format as specified in RFC4716
    - DSA keys are not supported. Make sure your key generator is set up to create RSA keys.
    Supported lengths: 1024, 2048, and 4096.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.import mykey publickey
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    try:
        key = conn.import_key_pair(key_name, public_key_material)
        log.debug("the key to return is : %s", key)
        return key.fingerprint
    except boto.exception.BotoServerError as e:
        log.debug(e)
        return False


def delete_key(key_name, region=None, key=None, keyid=None, profile=None):
    """
    Deletes a key. Always returns True

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.delete_key mykey
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    try:
        key = conn.delete_key_pair(key_name)
        log.debug("the key to return is : %s", key)
        return key
    except boto.exception.BotoServerError as e:
        log.debug(e)
        return False


def get_keys(
    keynames=None, filters=None, region=None, key=None, keyid=None, profile=None
):
    """
    Gets all keys or filters them by name and returns a list.
    keynames (list):: A list of the names of keypairs to retrieve.
    If not provided, all key pairs will be returned.
    filters (dict) :: Optional filters that can be used to limit the
    results returned. Filters are provided in the form of a dictionary
    consisting of filter names as the key and filter values as the
    value. The set of allowable filter names/values is dependent on
    the request being performed. Check the EC2 API guide for details.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.get_keys
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    try:
        keys = conn.get_all_key_pairs(keynames, filters)
        log.debug("the key to return is : %s", keys)
        key_values = []
        if keys:
            for key in keys:
                key_values.append(key.name)
        return key_values
    except boto.exception.BotoServerError as e:
        log.debug(e)
        return False


def get_attribute(
    attribute,
    instance_name=None,
    instance_id=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
    filters=None,
):
    """
    Get an EC2 instance attribute.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.get_attribute sourceDestCheck instance_name=my_instance

    Available attributes:
        * instanceType
        * kernel
        * ramdisk
        * userData
        * disableApiTermination
        * instanceInitiatedShutdownBehavior
        * rootDeviceName
        * blockDeviceMapping
        * productCodes
        * sourceDestCheck
        * groupSet
        * ebsOptimized
        * sriovNetSupport
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    attribute_list = [
        "instanceType",
        "kernel",
        "ramdisk",
        "userData",
        "disableApiTermination",
        "instanceInitiatedShutdownBehavior",
        "rootDeviceName",
        "blockDeviceMapping",
        "productCodes",
        "sourceDestCheck",
        "groupSet",
        "ebsOptimized",
        "sriovNetSupport",
    ]
    if not any((instance_name, instance_id)):
        raise SaltInvocationError(
            "At least one of the following must be specified: "
            "instance_name or instance_id."
        )
    if instance_name and instance_id:
        raise SaltInvocationError(
            "Both instance_name and instance_id can not be specified in the same"
            " command."
        )
    if attribute not in attribute_list:
        raise SaltInvocationError(
            "Attribute must be one of: {}.".format(attribute_list)
        )
    try:
        if instance_name:
            instances = find_instances(
                name=instance_name,
                region=region,
                key=key,
                keyid=keyid,
                profile=profile,
                filters=filters,
            )
            if len(instances) > 1:
                log.error("Found more than one EC2 instance matching the criteria.")
                return False
            elif not instances:
                log.error("Found no EC2 instance matching the criteria.")
                return False
            instance_id = instances[0]
        instance_attribute = conn.get_instance_attribute(instance_id, attribute)
        if not instance_attribute:
            return False
        return {attribute: instance_attribute[attribute]}
    except boto.exception.BotoServerError as exc:
        log.error(exc)
        return False


def set_attribute(
    attribute,
    attribute_value,
    instance_name=None,
    instance_id=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
    filters=None,
):
    """
    Set an EC2 instance attribute.
    Returns whether the operation succeeded or not.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.set_attribute sourceDestCheck False instance_name=my_instance

    Available attributes:
        * instanceType
        * kernel
        * ramdisk
        * userData
        * disableApiTermination
        * instanceInitiatedShutdownBehavior
        * rootDeviceName
        * blockDeviceMapping
        * productCodes
        * sourceDestCheck
        * groupSet
        * ebsOptimized
        * sriovNetSupport
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    attribute_list = [
        "instanceType",
        "kernel",
        "ramdisk",
        "userData",
        "disableApiTermination",
        "instanceInitiatedShutdownBehavior",
        "rootDeviceName",
        "blockDeviceMapping",
        "productCodes",
        "sourceDestCheck",
        "groupSet",
        "ebsOptimized",
        "sriovNetSupport",
    ]
    if not any((instance_name, instance_id)):
        raise SaltInvocationError(
            "At least one of the following must be specified: instance_name or"
            " instance_id."
        )
    if instance_name and instance_id:
        raise SaltInvocationError(
            "Both instance_name and instance_id can not be specified in the same"
            " command."
        )
    if attribute not in attribute_list:
        raise SaltInvocationError(
            "Attribute must be one of: {}.".format(attribute_list)
        )
    try:
        if instance_name:
            instances = find_instances(
                name=instance_name,
                region=region,
                key=key,
                keyid=keyid,
                profile=profile,
                filters=filters,
            )
            if len(instances) != 1:
                raise CommandExecutionError(
                    "Found more than one EC2 instance matching the criteria."
                )
            instance_id = instances[0]
        attribute = conn.modify_instance_attribute(
            instance_id, attribute, attribute_value
        )
        if not attribute:
            return False
        return attribute
    except boto.exception.BotoServerError as exc:
        log.error(exc)
        return False


def get_network_interface_id(name, region=None, key=None, keyid=None, profile=None):
    """
    Get an Elastic Network Interface id from its name tag.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.get_network_interface_id name=my_eni
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    r = {}
    try:
        enis = conn.get_all_network_interfaces(filters={"tag:Name": name})
        if not enis:
            r["error"] = {"message": "No ENIs found."}
        elif len(enis) > 1:
            r["error"] = {"message": "Name specified is tagged on multiple ENIs."}
        else:
            eni = enis[0]
            r["result"] = eni.id
    except boto.exception.EC2ResponseError as e:
        r["error"] = __utils__["boto.get_error"](e)
    return r


def get_network_interface(
    name=None,
    network_interface_id=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Get an Elastic Network Interface.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.get_network_interface name=my_eni
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    r = {}
    result = _get_network_interface(conn, name, network_interface_id)
    if "error" in result:
        if result["error"]["message"] == "No ENIs found.":
            r["result"] = None
            return r
        return result
    eni = result["result"]
    r["result"] = _describe_network_interface(eni)
    return r


def _get_network_interface(conn, name=None, network_interface_id=None):
    r = {}
    if not (name or network_interface_id):
        raise SaltInvocationError(
            "Either name or network_interface_id must be provided."
        )
    try:
        if network_interface_id:
            enis = conn.get_all_network_interfaces([network_interface_id])
        else:
            enis = conn.get_all_network_interfaces(filters={"tag:Name": name})

        if not enis:
            r["error"] = {"message": "No ENIs found."}
        elif len(enis) > 1:
            r["error"] = {"message": "Name specified is tagged on multiple ENIs."}
        else:
            eni = enis[0]
            r["result"] = eni
    except boto.exception.EC2ResponseError as e:
        r["error"] = __utils__["boto.get_error"](e)
    return r


def _describe_network_interface(eni):
    r = {}
    for attr in [
        "status",
        "description",
        "availability_zone",
        "requesterId",
        "requester_managed",
        "mac_address",
        "private_ip_address",
        "vpc_id",
        "id",
        "source_dest_check",
        "owner_id",
        "tags",
        "subnet_id",
        "associationId",
        "publicDnsName",
        "owner_id",
        "ipOwnerId",
        "publicIp",
        "allocationId",
    ]:
        if hasattr(eni, attr):
            r[attr] = getattr(eni, attr)
    r["region"] = eni.region.name
    r["groups"] = []
    for group in eni.groups:
        r["groups"].append({"name": group.name, "id": group.id})
    r["private_ip_addresses"] = []
    for address in eni.private_ip_addresses:
        r["private_ip_addresses"].append(
            {
                "private_ip_address": address.private_ip_address,
                "primary": address.primary,
            }
        )
    r["attachment"] = {}
    for attr in [
        "status",
        "attach_time",
        "device_index",
        "delete_on_termination",
        "instance_id",
        "instance_owner_id",
        "id",
    ]:
        if hasattr(eni.attachment, attr):
            r["attachment"][attr] = getattr(eni.attachment, attr)
    return r


def create_network_interface(
    name,
    subnet_id=None,
    subnet_name=None,
    private_ip_address=None,
    description=None,
    groups=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Create an Elastic Network Interface.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.create_network_interface my_eni subnet-12345 description=my_eni groups=['my_group']
    """
    if not salt.utils.data.exactly_one((subnet_id, subnet_name)):
        raise SaltInvocationError(
            "One (but not both) of subnet_id or subnet_name must be provided."
        )

    if subnet_name:
        resource = __salt__["boto_vpc.get_resource_id"](
            "subnet", subnet_name, region=region, key=key, keyid=keyid, profile=profile
        )
        if "id" not in resource:
            log.warning("Couldn't resolve subnet name %s.", subnet_name)
            return False
        subnet_id = resource["id"]

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    r = {}
    result = _get_network_interface(conn, name)
    if "result" in result:
        r["error"] = {"message": "An ENI with this Name tag already exists."}
        return r
    vpc_id = __salt__["boto_vpc.get_subnet_association"](
        [subnet_id], region=region, key=key, keyid=keyid, profile=profile
    )
    vpc_id = vpc_id.get("vpc_id")
    if not vpc_id:
        msg = "subnet_id {} does not map to a valid vpc id.".format(subnet_id)
        r["error"] = {"message": msg}
        return r
    _groups = __salt__["boto_secgroup.convert_to_group_ids"](
        groups, vpc_id=vpc_id, region=region, key=key, keyid=keyid, profile=profile
    )
    try:
        eni = conn.create_network_interface(
            subnet_id,
            private_ip_address=private_ip_address,
            description=description,
            groups=_groups,
        )
        eni.add_tag("Name", name)
    except boto.exception.EC2ResponseError as e:
        r["error"] = __utils__["boto.get_error"](e)
        return r
    r["result"] = _describe_network_interface(eni)
    return r


def delete_network_interface(
    name=None,
    network_interface_id=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Create an Elastic Network Interface.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.create_network_interface my_eni subnet-12345 description=my_eni groups=['my_group']
    """
    if not (name or network_interface_id):
        raise SaltInvocationError(
            "Either name or network_interface_id must be provided."
        )
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    r = {}
    result = _get_network_interface(conn, name, network_interface_id)
    if "error" in result:
        return result
    eni = result["result"]
    try:
        info = _describe_network_interface(eni)
        network_interface_id = info["id"]
    except KeyError:
        r["error"] = {"message": "ID not found for this network interface."}
        return r
    try:
        r["result"] = conn.delete_network_interface(network_interface_id)
    except boto.exception.EC2ResponseError as e:
        r["error"] = __utils__["boto.get_error"](e)
    return r


def attach_network_interface(
    device_index,
    name=None,
    network_interface_id=None,
    instance_name=None,
    instance_id=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Attach an Elastic Network Interface.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.attach_network_interface my_eni instance_name=salt-master device_index=0
    """
    if not salt.utils.data.exactly_one((name, network_interface_id)):
        raise SaltInvocationError(
            "Exactly one (but not both) of 'name' or 'network_interface_id' "
            "must be provided."
        )

    if not salt.utils.data.exactly_one((instance_name, instance_id)):
        raise SaltInvocationError(
            "Exactly one (but not both) of 'instance_name' or 'instance_id' "
            "must be provided."
        )

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    r = {}
    result = _get_network_interface(conn, name, network_interface_id)
    if "error" in result:
        return result
    eni = result["result"]
    try:
        info = _describe_network_interface(eni)
        network_interface_id = info["id"]
    except KeyError:
        r["error"] = {"message": "ID not found for this network interface."}
        return r

    if instance_name:
        try:
            instance_id = get_id(
                name=instance_name, region=region, key=key, keyid=keyid, profile=profile
            )
        except boto.exception.BotoServerError as e:
            log.error(e)
            return False

    try:
        r["result"] = conn.attach_network_interface(
            network_interface_id, instance_id, device_index
        )
    except boto.exception.EC2ResponseError as e:
        r["error"] = __utils__["boto.get_error"](e)
    return r


def detach_network_interface(
    name=None,
    network_interface_id=None,
    attachment_id=None,
    force=False,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Detach an Elastic Network Interface.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.detach_network_interface my_eni
    """
    if not (name or network_interface_id or attachment_id):
        raise SaltInvocationError(
            "Either name or network_interface_id or attachment_id must be provided."
        )
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    r = {}
    if not attachment_id:
        result = _get_network_interface(conn, name, network_interface_id)
        if "error" in result:
            return result
        eni = result["result"]
        info = _describe_network_interface(eni)
        try:
            attachment_id = info["attachment"]["id"]
        except KeyError:
            r["error"] = {"message": "Attachment id not found for this ENI."}
            return r
    try:
        r["result"] = conn.detach_network_interface(attachment_id, force)
    except boto.exception.EC2ResponseError as e:
        r["error"] = __utils__["boto.get_error"](e)
    return r


def modify_network_interface_attribute(
    name=None,
    network_interface_id=None,
    attr=None,
    value=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Modify an attribute of an Elastic Network Interface.

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt myminion boto_ec2.modify_network_interface_attribute my_eni attr=description value='example description'
    """
    if not (name or network_interface_id):
        raise SaltInvocationError(
            "Either name or network_interface_id must be provided."
        )
    if attr is None and value is None:
        raise SaltInvocationError("attr and value must be provided.")
    r = {}
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    result = _get_network_interface(conn, name, network_interface_id)
    if "error" in result:
        return result
    eni = result["result"]
    info = _describe_network_interface(eni)
    network_interface_id = info["id"]
    # munge attr into what the API requires
    if attr == "groups":
        _attr = "groupSet"
    elif attr == "source_dest_check":
        _attr = "sourceDestCheck"
    elif attr == "delete_on_termination":
        _attr = "deleteOnTermination"
    else:
        _attr = attr
    _value = value
    if info.get("vpc_id") and _attr == "groupSet":
        _value = __salt__["boto_secgroup.convert_to_group_ids"](
            value,
            vpc_id=info.get("vpc_id"),
            region=region,
            key=key,
            keyid=keyid,
            profile=profile,
        )
        if not _value:
            r["error"] = {
                "message": "Security groups do not map to valid security group ids"
            }
            return r
    _attachment_id = None
    if _attr == "deleteOnTermination":
        try:
            _attachment_id = info["attachment"]["id"]
        except KeyError:
            r["error"] = {
                "message": (
                    "No attachment id found for this ENI. The ENI must"
                    " be attached before delete_on_termination can be"
                    " modified"
                )
            }
            return r
    try:
        r["result"] = conn.modify_network_interface_attribute(
            network_interface_id, _attr, _value, attachment_id=_attachment_id
        )
    except boto.exception.EC2ResponseError as e:
        r["error"] = __utils__["boto.get_error"](e)
    return r


def get_all_volumes(
    volume_ids=None,
    filters=None,
    return_objs=False,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Get a list of all EBS volumes, optionally filtered by provided 'filters' param

    .. versionadded:: 2016.11.0

    volume_ids
        (list) - Optional list of volume_ids.  If provided, only the volumes
        associated with those in the list will be returned.
    filters
        (dict) - Additional constraints on which volumes to return.  Valid filters are:

    - attachment.attach-time - The time stamp when the attachment initiated.
    - attachment.delete-on-termination - Whether the volume is deleted on instance termination.
    - attachment.device - The device name that is exposed to the instance (for example, /dev/sda1).
    - attachment.instance-id - The ID of the instance the volume is attached to.
    - attachment.status - The attachment state (attaching | attached | detaching | detached).
    - availability-zone - The Availability Zone in which the volume was created.
    - create-time - The time stamp when the volume was created.
    - encrypted - The encryption status of the volume.
    - size - The size of the volume, in GiB.
    - snapshot-id - The snapshot from which the volume was created.
    - status - The status of the volume (creating | available | in-use | deleting | deleted | error).
    - tag:key=value - The key/value combination of a tag assigned to the resource.
    - volume-id - The volume ID.
    - volume-type - The Amazon EBS volume type. This can be ``gp2`` for General
      Purpose SSD, ``io1`` for Provisioned IOPS SSD, ``st1`` for Throughput
      Optimized HDD, ``sc1`` for Cold HDD, or ``standard`` for Magnetic volumes.

    return_objs
        (bool) - Changes the return type from list of volume IDs to list of
        boto.ec2.volume.Volume objects

    returns
        (list) - A list of the requested values: Either the volume IDs or, if
        return_objs is ``True``, boto.ec2.volume.Volume objects.

    CLI Example:

    .. code-block:: bash

        salt-call boto_ec2.get_all_volumes filters='{"tag:Name": "myVolume01"}'

    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)

    try:
        ret = conn.get_all_volumes(volume_ids=volume_ids, filters=filters)
        return ret if return_objs else [r.id for r in ret]
    except boto.exception.BotoServerError as e:
        log.error(e)
        return []


def set_volumes_tags(
    tag_maps,
    authoritative=False,
    dry_run=False,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    .. versionadded:: 2016.11.0

    tag_maps (list)
        List of dicts of filters and tags, where 'filters' is a dict suitable for passing to the
        'filters' argument of get_all_volumes() above, and 'tags' is a dict of tags to be set on
        volumes (via create_tags/delete_tags) as matched by the given filters.  The filter syntax
        is extended to permit passing either a list of volume_ids or an instance_name (with
        instance_name being the Name tag of the instance to which the desired volumes are mapped).
        Each mapping in the list is applied separately, so multiple sets of volumes can be all
        tagged differently with one call to this function.  If filtering by instance Name, You may
        additionally limit the instances matched by passing in a list of desired instance states.
        The default set of states is ('pending', 'rebooting', 'running', 'stopping', 'stopped').

    YAML example fragment:

    .. code-block:: yaml

        - filters:
            attachment.instance_id: i-abcdef12
          tags:
            Name: dev-int-abcdef12.aws-foo.com
        - filters:
            attachment.device: /dev/sdf
          tags:
            ManagedSnapshots: true
            BillingGroup: bubba.hotep@aws-foo.com
          in_states:
          - stopped
          - terminated
        - filters:
            instance_name: prd-foo-01.aws-foo.com
          tags:
            Name: prd-foo-01.aws-foo.com
            BillingGroup: infra-team@aws-foo.com
        - filters:
            volume_ids: [ vol-12345689, vol-abcdef12 ]
          tags:
            BillingGroup: infra-team@aws-foo.com

    authoritative (bool)
        If true, any existing tags on the matched volumes, and not explicitly requested here, will
        be removed.

    dry_run (bool)
        If true, don't change anything, just return a dictionary describing any changes which
        would have been applied.

    returns (dict)
        A dict describing status and any changes.

    """
    ret = {"success": True, "comment": "", "changes": {}}
    running_states = ("pending", "rebooting", "running", "stopping", "stopped")
    ### First creeate a dictionary mapping all changes for a given volume to its volume ID...
    tag_sets = {}
    for tm in tag_maps:
        filters = dict(tm.get("filters", {}))
        tags = dict(tm.get("tags", {}))
        args = {
            "return_objs": True,
            "region": region,
            "key": key,
            "keyid": keyid,
            "profile": profile,
        }
        new_filters = {}
        log.debug("got filters: %s", filters)
        instance_id = None
        in_states = tm.get("in_states", running_states)
        try:
            for k, v in filters.items():
                if k == "volume_ids":
                    args["volume_ids"] = v
                elif k == "instance_name":
                    instance_id = get_id(
                        name=v,
                        in_states=in_states,
                        region=region,
                        key=key,
                        keyid=keyid,
                        profile=profile,
                    )
                    if not instance_id:
                        msg = "Couldn't resolve instance Name {} to an ID.".format(v)
                        raise CommandExecutionError(msg)
                    new_filters["attachment.instance_id"] = instance_id
                else:
                    new_filters[k] = v
        except CommandExecutionError as e:
            log.warning(e)
            continue  # Hmme, abort or do what we can...?  Guess the latter for now.
        args["filters"] = new_filters
        volumes = get_all_volumes(**args)
        log.debug("got volume list: %s", volumes)
        for vol in volumes:
            tag_sets.setdefault(
                vol.id.replace("-", "_"), {"vol": vol, "tags": tags.copy()}
            )["tags"].update(tags.copy())
    log.debug("tag_sets after munging: %s", tag_sets)

    ### ...then loop through all those volume->tag pairs and apply them.
    changes = {"old": {}, "new": {}}
    for volume in tag_sets.values():
        vol, tags = volume["vol"], volume["tags"]
        log.debug(
            "current tags on vol.id %s: %s", vol.id, dict(getattr(vol, "tags", {}))
        )
        curr = set(dict(getattr(vol, "tags", {})).keys())
        log.debug("requested tags on vol.id %s: %s", vol.id, tags)
        req = set(tags.keys())
        add = list(req - curr)
        update = [r for r in (req & curr) if vol.tags[r] != tags[r]]
        remove = list(curr - req)
        if add or update or (authoritative and remove):
            changes["old"][vol.id] = dict(getattr(vol, "tags", {}))
            changes["new"][vol.id] = tags
        else:
            log.debug("No changes needed for vol.id %s", vol.id)
        if add:
            d = {k: tags[k] for k in add}
            log.debug("New tags for vol.id %s: %s", vol.id, d)
        if update:
            d = {k: tags[k] for k in update}
            log.debug("Updated tags for vol.id %s: %s", vol.id, d)
        if not dry_run:
            if not create_tags(
                vol.id, tags, region=region, key=key, keyid=keyid, profile=profile
            ):
                ret["success"] = False
                ret["comment"] = "Failed to set tags on vol.id {}: {}".format(
                    vol.id, tags
                )
                return ret
            if authoritative:
                if remove:
                    log.debug("Removed tags for vol.id %s: %s", vol.id, remove)
                    if not delete_tags(
                        vol.id,
                        remove,
                        region=region,
                        key=key,
                        keyid=keyid,
                        profile=profile,
                    ):
                        ret["success"] = False
                        ret[
                            "comment"
                        ] = "Failed to remove tags on vol.id {}: {}".format(
                            vol.id, remove
                        )
                        return ret
    if changes["old"] or changes["new"]:
        ret["changes"].update(changes)
    return ret


def get_all_tags(filters=None, region=None, key=None, keyid=None, profile=None):
    """
    Describe all tags matching the filter criteria, or all tags in the account otherwise.

    .. versionadded:: 2018.3.0

    filters
        (dict) - Additional constraints on which volumes to return.  Note that valid filters vary
        extensively depending on the resource type.  When in doubt, search first without a filter
        and then use the returned data to help fine-tune your search.  You can generally garner the
        resource type from its ID (e.g. `vol-XXXXX` is a volume, `i-XXXXX` is an instance, etc.

    CLI Example:

    .. code-block:: bash

        salt-call boto_ec2.get_all_tags '{"tag:Name": myInstanceNameTag, resource-type: instance}'

    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    try:
        ret = conn.get_all_tags(filters)
        tags = {}
        for t in ret:
            if t.res_id not in tags:
                tags[t.res_id] = {}
            tags[t.res_id][t.name] = t.value
        return tags
    except boto.exception.BotoServerError as e:
        log.error(e)
        return {}


def create_tags(resource_ids, tags, region=None, key=None, keyid=None, profile=None):
    """
    Create new metadata tags for the specified resource ids.

    .. versionadded:: 2016.11.0

    resource_ids
        (string) or (list) – List of resource IDs.  A plain string will be converted to a list of one element.
    tags
        (dict) – Dictionary of name/value pairs. To create only a tag name, pass '' as the value.

    returns
        (bool) - True on success, False on failure.

    CLI Example:

    .. code-block:: bash

        salt-call boto_ec2.create_tags vol-12345678 '{"Name": "myVolume01"}'

    """
    if not isinstance(resource_ids, list):
        resource_ids = [resource_ids]

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    try:
        conn.create_tags(resource_ids, tags)
        return True
    except boto.exception.BotoServerError as e:
        log.error(e)
        return False


def delete_tags(resource_ids, tags, region=None, key=None, keyid=None, profile=None):
    """
    Delete metadata tags for the specified resource ids.

    .. versionadded:: 2016.11.0

    resource_ids
        (string) or (list) – List of resource IDs.  A plain string will be converted to a list of one element.
    tags
        (dict) or (list) – Either a dictionary containing name/value pairs or a list containing just tag names.
                           If you pass in a dictionary, the values must match the actual tag values or the tag
                           will not be deleted. If you pass in a value of None for the tag value, all tags with
                           that name will be deleted.

    returns
        (bool) - True on success, False on failure.

    CLI Example:

    .. code-block:: bash

        salt-call boto_ec2.delete_tags vol-12345678 '{"Name": "myVolume01"}'
        salt-call boto_ec2.delete_tags vol-12345678 '["Name","MountPoint"]'

    """
    if not isinstance(resource_ids, list):
        resource_ids = [resource_ids]

    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    try:
        conn.delete_tags(resource_ids, tags)
        return True
    except boto.exception.BotoServerError as e:
        log.error(e)
        return False


def detach_volume(
    volume_id,
    instance_id=None,
    device=None,
    force=False,
    wait_for_detachement=False,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Detach an EBS volume from an EC2 instance.

    .. versionadded:: 2016.11.0

    volume_id
        (string) – The ID of the EBS volume to be detached.
    instance_id
        (string) – The ID of the EC2 instance from which it will be detached.
    device
        (string) – The device on the instance through which the volume is exposted (e.g. /dev/sdh)
    force
        (bool) – Forces detachment if the previous detachment attempt did not occur cleanly.
                 This option can lead to data loss or a corrupted file system. Use this option
                 only as a last resort to detach a volume from a failed instance. The instance
                 will not have an opportunity to flush file system caches nor file system meta data.
                 If you use this option, you must perform file system check and repair procedures.
    wait_for_detachement
       (bool) - Whether or not to wait for volume detachement to complete.

    returns
        (bool) - True on success, False on failure.

    CLI Example:

    .. code-block:: bash

        salt-call boto_ec2.detach_volume vol-12345678 i-87654321

    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    try:
        ret = conn.detach_volume(volume_id, instance_id, device, force)
        if (
            ret
            and wait_for_detachement
            and not _wait_for_volume_available(conn, volume_id)
        ):
            timeout_msg = 'Timed out waiting for the volume status "available".'
            log.error(timeout_msg)
            return False
        return ret
    except boto.exception.BotoServerError as e:
        log.error(e)
        return False


def delete_volume(
    volume_id,
    instance_id=None,
    device=None,
    force=False,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Detach an EBS volume from an EC2 instance.

    .. versionadded:: 2016.11.0

    volume_id
        (string) – The ID of the EBS volume to be deleted.
    force
        (bool) – Forces deletion even if the device has not yet been detached from its instance.

    returns
        (bool) - True on success, False on failure.

    CLI Example:

    .. code-block:: bash

        salt-call boto_ec2.delete_volume vol-12345678

    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    try:
        return conn.delete_volume(volume_id)
    except boto.exception.BotoServerError as e:
        if not force:
            log.error(e)
            return False
    try:
        conn.detach_volume(volume_id, force=force)
        return conn.delete_volume(volume_id)
    except boto.exception.BotoServerError as e:
        log.error(e)
        return False


def _wait_for_volume_available(conn, volume_id, retries=5, interval=5):
    i = 0
    while True:
        i = i + 1
        time.sleep(interval)
        vols = conn.get_all_volumes(volume_ids=[volume_id])
        if len(vols) != 1:
            return False
        vol = vols[0]
        if vol.status == "available":
            return True
        if i > retries:
            return False


def attach_volume(
    volume_id, instance_id, device, region=None, key=None, keyid=None, profile=None
):
    """
    Attach an EBS volume to an EC2 instance.
    ..

    volume_id
        (string) – The ID of the EBS volume to be attached.
    instance_id
        (string) – The ID of the EC2 instance to attach the volume to.
    device
        (string) – The device on the instance through which the volume is exposed (e.g. /dev/sdh)

    returns
        (bool) - True on success, False on failure.

    CLI Example:

    .. code-block:: bash

        salt-call boto_ec2.attach_volume vol-12345678 i-87654321 /dev/sdh

    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    try:
        return conn.attach_volume(volume_id, instance_id, device)
    except boto.exception.BotoServerError as error:
        log.error(error)
        return False


def create_volume(
    zone_name,
    size=None,
    snapshot_id=None,
    volume_type=None,
    iops=None,
    encrypted=False,
    kms_key_id=None,
    wait_for_creation=False,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Create an EBS volume to an availability zone.

    ..

    zone_name
        (string) – The Availability zone name of the EBS volume to be created.
    size
        (int) –  The size of the new volume, in GiB. If you're creating the
                 volume from a snapshot and don't specify a volume size, the
                 default is the snapshot size.
    snapshot_id
        (string) –  The snapshot ID from which the new volume will be created.
    volume_type
        (string) -  The type of the volume. Valid volume types for AWS can be found here:
                    http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/EBSVolumeTypes.html
    iops
        (int) - The provisioned IOPS you want to associate with this volume.
    encrypted
        (bool) - Specifies whether the volume should be encrypted.
    kms_key_id
        (string) - If encrypted is True, this KMS Key ID may be specified to
                   encrypt volume with this key
                   e.g.: arn:aws:kms:us-east-1:012345678910:key/abcd1234-a123-456a-a12b-a123b4cd56ef
    wait_for_creation
        (bool) - Whether or not to wait for volume creation to complete.

    returns
        (string) - created volume id on success, error message on failure.

    CLI Example:

    .. code-block:: bash

        salt-call boto_ec2.create_volume us-east-1a size=10
        salt-call boto_ec2.create_volume us-east-1a snapshot_id=snap-0123abcd

    """
    if size is None and snapshot_id is None:
        raise SaltInvocationError("Size must be provided if not created from snapshot.")
    ret = {}
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    try:
        vol = conn.create_volume(
            size=size,
            zone=zone_name,
            snapshot=snapshot_id,
            volume_type=volume_type,
            iops=iops,
            encrypted=encrypted,
            kms_key_id=kms_key_id,
        )
        if wait_for_creation and not _wait_for_volume_available(conn, vol.id):
            timeout_msg = 'Timed out waiting for the volume status "available".'
            log.error(timeout_msg)
            ret["error"] = timeout_msg
        else:
            ret["result"] = vol.id
    except boto.exception.BotoServerError as error:
        ret["error"] = __utils__["boto.get_error"](error)
    return ret



import logging

import salt.pillar.s3 as s3_pillar
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, mock_open, patch
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class S3PillarTestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for salt.pillar.s3
    """

    def setup_loader_modules(self):
        s3_pillar_globals = {"__utils__": {}}
        return {s3_pillar: s3_pillar_globals}

    def test_refresh_buckets_cache_file(self):
        """
        Test pagination with refresh_buckets_cache_file
        """
        key = "XXXXXXXXXXXXXXXXXXXXX"
        keyid = "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"
        bucket = "dummy_bucket"
        service_url = "s3.amazonaws.com"
        cache_file = "dummy_file"

        s3_creds = s3_pillar.S3Credentials(key, keyid, bucket, service_url)

        mock_return_first = [
            {"Name": "pillar-bucket"},
            {"Prefix": "test"},
            {"KeyCount": "10"},
            {"MaxKeys": "10"},
            {"NextContinuationToken": "XXXXX"},
            {"IsTruncated": "true"},
        ]

        mock_return_second = [
            {"Name": "pillar-bucket"},
            {"Prefix": "test"},
            {"KeyCount": "10"},
            {"MaxKeys": "10"},
            {"IsTruncated": "true"},
        ]

        first_range_end = 999
        second_range_end = 1200
        for i in range(0, first_range_end):
            key_name = "{}/init.sls".format(i)
            tmp = {
                "Key": key_name,
                "LastModified": "2019-12-18T15:54:39.000Z",
                "ETag": '"fba0a053704e8b357c94be90b44bb640"',
                "Size": "5 ",
                "StorageClass": "STANDARD",
            }
            mock_return_first.append(tmp)

        for i in range(first_range_end, second_range_end):
            key_name = "{}/init.sls".format(i)
            tmp = {
                "Key": key_name,
                "LastModified": "2019-12-18T15:54:39.000Z",
                "ETag": '"fba0a053704e8b357c94be90b44bb640"',
                "Size": "5 ",
                "StorageClass": "STANDARD",
            }
            mock_return_second.append(tmp)

        _expected = {"base": {"dummy_bucket": []}}
        for i in range(0, second_range_end):
            key_name = "{}/init.sls".format(i)
            tmp = {
                "Key": key_name,
                "LastModified": "2019-12-18T15:54:39.000Z",
                "ETag": '"fba0a053704e8b357c94be90b44bb640"',
                "Size": "5 ",
                "StorageClass": "STANDARD",
            }
            _expected["base"]["dummy_bucket"].append(tmp)

        mock_s3_query = MagicMock(side_effect=[mock_return_first, mock_return_second])
        with patch.dict(s3_pillar.__utils__, {"s3.query": mock_s3_query}):
            with patch("salt.utils.files.fopen", mock_open(read_data=b"")):
                ret = s3_pillar._refresh_buckets_cache_file(
                    s3_creds, cache_file, False, "base", ""
                )
                self.assertEqual(ret, _expected)



"""
Zeromq transport classes
"""
import errno
import hashlib
import logging
import os
import signal
import sys
import threading
import uuid
from random import randint

import salt.auth
import salt.crypt
import salt.ext.tornado
import salt.ext.tornado.concurrent
import salt.ext.tornado.gen
import salt.ext.tornado.ioloop
import salt.log.setup
import salt.payload
import salt.transport.client
import salt.transport.mixins.auth
import salt.transport.server
import salt.utils.event
import salt.utils.files
import salt.utils.minions
import salt.utils.process
import salt.utils.stringutils
import salt.utils.verify
import salt.utils.versions
import salt.utils.zeromq
import zmq.error
import zmq.eventloop.ioloop
import zmq.eventloop.zmqstream
from salt._compat import ipaddress
from salt.exceptions import SaltException, SaltReqTimeoutError
from salt.utils.zeromq import LIBZMQ_VERSION_INFO, ZMQ_VERSION_INFO, zmq

try:
    import zmq.utils.monitor

    HAS_ZMQ_MONITOR = True
except ImportError:
    HAS_ZMQ_MONITOR = False


try:
    from M2Crypto import RSA

    HAS_M2 = True
except ImportError:
    HAS_M2 = False
    try:
        from Cryptodome.Cipher import PKCS1_OAEP
    except ImportError:
        from Crypto.Cipher import PKCS1_OAEP  # nosec


log = logging.getLogger(__name__)


def _get_master_uri(master_ip, master_port, source_ip=None, source_port=None):
    """
    Return the ZeroMQ URI to connect the Minion to the Master.
    It supports different source IP / port, given the ZeroMQ syntax:
    // Connecting using a IP address and bind to an IP address
    rc = zmq_connect(socket, "tcp://192.168.1.17:5555;192.168.1.1:5555"); assert (rc == 0);
    Source: http://api.zeromq.org/4-1:zmq-tcp
    """

    from salt.utils.zeromq import ip_bracket

    master_uri = "tcp://{master_ip}:{master_port}".format(
        master_ip=ip_bracket(master_ip), master_port=master_port
    )
    if source_ip or source_port:
        if LIBZMQ_VERSION_INFO >= (4, 1, 6) and ZMQ_VERSION_INFO >= (16, 0, 1):
            # The source:port syntax for ZeroMQ has been added in libzmq 4.1.6
            # which is included in the pyzmq wheels starting with 16.0.1.
            if source_ip and source_port:
                master_uri = (
                    "tcp://{source_ip}:{source_port};{master_ip}:{master_port}".format(
                        source_ip=ip_bracket(source_ip),
                        source_port=source_port,
                        master_ip=ip_bracket(master_ip),
                        master_port=master_port,
                    )
                )
            elif source_ip and not source_port:
                master_uri = "tcp://{source_ip}:0;{master_ip}:{master_port}".format(
                    source_ip=ip_bracket(source_ip),
                    master_ip=ip_bracket(master_ip),
                    master_port=master_port,
                )
            elif source_port and not source_ip:
                ip_any = (
                    "0.0.0.0"
                    if ipaddress.ip_address(master_ip).version == 4
                    else ip_bracket("::")
                )
                master_uri = (
                    "tcp://{ip_any}:{source_port};{master_ip}:{master_port}".format(
                        ip_any=ip_any,
                        source_port=source_port,
                        master_ip=ip_bracket(master_ip),
                        master_port=master_port,
                    )
                )
        else:
            log.warning(
                "Unable to connect to the Master using a specific source IP / port"
            )
            log.warning("Consider upgrading to pyzmq >= 16.0.1 and libzmq >= 4.1.6")
            log.warning(
                "Specific source IP / port for connecting to master returner port:"
                " configuraion ignored"
            )

    return master_uri


class AsyncZeroMQReqChannel(salt.transport.client.ReqChannel):
    """
    Encapsulate sending routines to ZeroMQ.

    ZMQ Channels default to 'crypt=aes'
    """

    async_methods = [
        "crypted_transfer_decode_dictentry",
        "_crypted_transfer",
        "_do_transfer",
        "_uncrypted_transfer",
        "send",
    ]
    close_methods = [
        "close",
    ]

    def __init__(self, opts, **kwargs):
        self.opts = dict(opts)
        self.ttype = "zeromq"

        # crypt defaults to 'aes'
        self.crypt = kwargs.get("crypt", "aes")

        if "master_uri" in kwargs:
            self.opts["master_uri"] = kwargs["master_uri"]

        self._io_loop = kwargs.get("io_loop")
        if self._io_loop is None:
            self._io_loop = salt.ext.tornado.ioloop.IOLoop.current()

        if self.crypt != "clear":
            # we don't need to worry about auth as a kwarg, since its a singleton
            self.auth = salt.crypt.AsyncAuth(self.opts, io_loop=self._io_loop)
        log.debug(
            "Connecting the Minion to the Master URI (for the return server): %s",
            self.master_uri,
        )
        self.message_client = AsyncReqMessageClientPool(
            self.opts,
            args=(
                self.opts,
                self.master_uri,
            ),
            kwargs={"io_loop": self._io_loop},
        )
        self._closing = False

    def close(self):
        """
        Since the message_client creates sockets and assigns them to the IOLoop we have to
        specifically destroy them, since we aren't the only ones with references to the FDs
        """
        if self._closing:
            return
        log.debug("Closing %s instance", self.__class__.__name__)
        self._closing = True
        if hasattr(self, "message_client"):
            self.message_client.close()

    # pylint: disable=W1701
    def __del__(self):
        try:
            self.close()
        except OSError as exc:
            if exc.errno != errno.EBADF:
                # If its not a bad file descriptor error, raise
                raise

    # pylint: enable=W1701

    @property
    def master_uri(self):
        if "master_uri" in self.opts:
            return self.opts["master_uri"]

        # if by chance master_uri is not there..
        if "master_ip" in self.opts:
            return _get_master_uri(
                self.opts["master_ip"],
                self.opts["master_port"],
                source_ip=self.opts.get("source_ip"),
                source_port=self.opts.get("source_ret_port"),
            )

        # if we've reached here something is very abnormal
        raise SaltException("ReqChannel: missing master_uri/master_ip in self.opts")

    def _package_load(self, load):
        return {
            "enc": self.crypt,
            "load": load,
            "version": 2,
        }

    @salt.ext.tornado.gen.coroutine
    def crypted_transfer_decode_dictentry(
        self, load, dictkey=None, tries=3, timeout=60
    ):
        nonce = uuid.uuid4().hex
        load["nonce"] = nonce
        if not self.auth.authenticated:
            # Return control back to the caller, continue when authentication succeeds
            yield self.auth.authenticate()

        # Return control to the caller. When send() completes, resume by
        # populating ret with the Future.result
        ret = yield self.message_client.send(
            self._package_load(self.auth.crypticle.dumps(load)),
            timeout=timeout,
            tries=tries,
        )

        if "key" not in ret:
            # Reauth in the case our key is deleted on the master side.
            yield self.auth.authenticate()
            ret = yield self.message_client.send(
                self._package_load(self.auth.crypticle.dumps(load)),
                timeout=timeout,
                tries=tries,
            )

        key = self.auth.get_keys()
        if HAS_M2:
            aes = key.private_decrypt(ret["key"], RSA.pkcs1_oaep_padding)
        else:
            cipher = PKCS1_OAEP.new(key)
            aes = cipher.decrypt(ret["key"])

        # Decrypt using the public key.
        pcrypt = salt.crypt.Crypticle(self.opts, aes)
        signed_msg = pcrypt.loads(ret[dictkey])

        # Validate the master's signature.
        master_pubkey_path = os.path.join(self.opts["pki_dir"], "minion_master.pub")
        if not salt.crypt.verify_signature(
            master_pubkey_path, signed_msg["data"], signed_msg["sig"]
        ):
            raise salt.crypt.AuthenticationError(
                "Pillar payload signature failed to validate."
            )

        # Make sure the signed key matches the key we used to decrypt the data.
        data = salt.payload.loads(signed_msg["data"])
        if data["key"] != ret["key"]:
            raise salt.crypt.AuthenticationError("Key verification failed.")

        # Validate the nonce.
        if data["nonce"] != nonce:
            raise salt.crypt.AuthenticationError("Pillar nonce verification failed.")
        raise salt.ext.tornado.gen.Return(data["pillar"])

    @salt.ext.tornado.gen.coroutine
    def _crypted_transfer(self, load, tries=3, timeout=60, raw=False):
        """
        Send a load across the wire, with encryption

        In case of authentication errors, try to renegotiate authentication
        and retry the method.

        Indeed, we can fail too early in case of a master restart during a
        minion state execution call

        :param dict load: A load to send across the wire
        :param int tries: The number of times to make before failure
        :param int timeout: The number of seconds on a response before failing
        """
        nonce = uuid.uuid4().hex
        if load and isinstance(load, dict):
            load["nonce"] = nonce

        @salt.ext.tornado.gen.coroutine
        def _do_transfer():
            # Yield control to the caller. When send() completes, resume by populating data with the Future.result
            data = yield self.message_client.send(
                self._package_load(self.auth.crypticle.dumps(load)),
                timeout=timeout,
                tries=tries,
            )
            # we may not have always data
            # as for example for saltcall ret submission, this is a blind
            # communication, we do not subscribe to return events, we just
            # upload the results to the master
            if data:
                data = self.auth.crypticle.loads(data, raw, nonce)
            if not raw:
                data = salt.transport.frame.decode_embedded_strs(data)
            raise salt.ext.tornado.gen.Return(data)

        if not self.auth.authenticated:
            # Return control back to the caller, resume when authentication succeeds
            yield self.auth.authenticate()
        try:
            # We did not get data back the first time. Retry.
            ret = yield _do_transfer()
        except salt.crypt.AuthenticationError:
            # If auth error, return control back to the caller, continue when authentication succeeds
            yield self.auth.authenticate()
            ret = yield _do_transfer()
        raise salt.ext.tornado.gen.Return(ret)

    @salt.ext.tornado.gen.coroutine
    def _uncrypted_transfer(self, load, tries=3, timeout=60):
        """
        Send a load across the wire in cleartext

        :param dict load: A load to send across the wire
        :param int tries: The number of times to make before failure
        :param int timeout: The number of seconds on a response before failing
        """
        ret = yield self.message_client.send(
            self._package_load(load),
            timeout=timeout,
            tries=tries,
        )

        raise salt.ext.tornado.gen.Return(ret)

    @salt.ext.tornado.gen.coroutine
    def send(self, load, tries=3, timeout=60, raw=False):
        """
        Send a request, return a future which will complete when we send the message
        """
        if self.crypt == "clear":
            ret = yield self._uncrypted_transfer(load, tries=tries, timeout=timeout)
        else:
            ret = yield self._crypted_transfer(
                load, tries=tries, timeout=timeout, raw=raw
            )
        raise salt.ext.tornado.gen.Return(ret)


class AsyncZeroMQPubChannel(
    salt.transport.mixins.auth.AESPubClientMixin, salt.transport.client.AsyncPubChannel
):
    """
    A transport channel backed by ZeroMQ for a Salt Publisher to use to
    publish commands to connected minions
    """

    async_methods = [
        "connect",
        "_decode_messages",
    ]
    close_methods = [
        "close",
    ]

    def __init__(self, opts, **kwargs):
        self.opts = opts
        self.ttype = "zeromq"
        self.io_loop = kwargs.get("io_loop")
        self._closing = False

        if self.io_loop is None:
            self.io_loop = salt.ext.tornado.ioloop.IOLoop.current()

        self.hexid = hashlib.sha1(
            salt.utils.stringutils.to_bytes(self.opts["id"])
        ).hexdigest()
        self.auth = salt.crypt.AsyncAuth(self.opts, io_loop=self.io_loop)
        self.context = zmq.Context()
        self._socket = self.context.socket(zmq.SUB)

        if self.opts["zmq_filtering"]:
            # TODO: constants file for "broadcast"
            self._socket.setsockopt(zmq.SUBSCRIBE, b"broadcast")
            if self.opts.get("__role") == "syndic":
                self._socket.setsockopt(zmq.SUBSCRIBE, b"syndic")
            else:
                self._socket.setsockopt(
                    zmq.SUBSCRIBE, salt.utils.stringutils.to_bytes(self.hexid)
                )
        else:
            self._socket.setsockopt(zmq.SUBSCRIBE, b"")

        self._socket.setsockopt(
            zmq.IDENTITY, salt.utils.stringutils.to_bytes(self.opts["id"])
        )

        # TODO: cleanup all the socket opts stuff
        if hasattr(zmq, "TCP_KEEPALIVE"):
            self._socket.setsockopt(zmq.TCP_KEEPALIVE, self.opts["tcp_keepalive"])
            self._socket.setsockopt(
                zmq.TCP_KEEPALIVE_IDLE, self.opts["tcp_keepalive_idle"]
            )
            self._socket.setsockopt(
                zmq.TCP_KEEPALIVE_CNT, self.opts["tcp_keepalive_cnt"]
            )
            self._socket.setsockopt(
                zmq.TCP_KEEPALIVE_INTVL, self.opts["tcp_keepalive_intvl"]
            )

        recon_delay = self.opts["recon_default"]

        if self.opts["recon_randomize"]:
            recon_delay = randint(
                self.opts["recon_default"],
                self.opts["recon_default"] + self.opts["recon_max"],
            )

            log.debug(
                "Generated random reconnect delay between '%sms' and '%sms' (%s)",
                self.opts["recon_default"],
                self.opts["recon_default"] + self.opts["recon_max"],
                recon_delay,
            )

        log.debug("Setting zmq_reconnect_ivl to '%sms'", recon_delay)
        self._socket.setsockopt(zmq.RECONNECT_IVL, recon_delay)

        if hasattr(zmq, "RECONNECT_IVL_MAX"):
            log.debug(
                "Setting zmq_reconnect_ivl_max to '%sms'",
                self.opts["recon_default"] + self.opts["recon_max"],
            )

            self._socket.setsockopt(zmq.RECONNECT_IVL_MAX, self.opts["recon_max"])

        if (self.opts["ipv6"] is True or ":" in self.opts["master_ip"]) and hasattr(
            zmq, "IPV4ONLY"
        ):
            # IPv6 sockets work for both IPv6 and IPv4 addresses
            self._socket.setsockopt(zmq.IPV4ONLY, 0)

        if HAS_ZMQ_MONITOR and self.opts["zmq_monitor"]:
            self._monitor = ZeroMQSocketMonitor(self._socket)
            self._monitor.start_io_loop(self.io_loop)

    def close(self):
        if self._closing is True:
            return

        self._closing = True

        if hasattr(self, "_monitor") and self._monitor is not None:
            self._monitor.stop()
            self._monitor = None
        if hasattr(self, "_stream"):
            self._stream.close(0)
        elif hasattr(self, "_socket"):
            self._socket.close(0)
        if hasattr(self, "context") and self.context.closed is False:
            self.context.term()

    # pylint: disable=W1701
    def __del__(self):
        self.close()

    # pylint: enable=W1701
    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    # TODO: this is the time to see if we are connected, maybe use the req channel to guess?
    @salt.ext.tornado.gen.coroutine
    def connect(self):
        if not self.auth.authenticated:
            yield self.auth.authenticate()

        # if this is changed from the default, we assume it was intentional
        if int(self.opts.get("publish_port", 4506)) != 4506:
            self.publish_port = self.opts.get("publish_port")
        # else take the relayed publish_port master reports
        else:
            self.publish_port = self.auth.creds["publish_port"]

        log.debug(
            "Connecting the Minion to the Master publish port, using the URI: %s",
            self.master_pub,
        )
        self._socket.connect(self.master_pub)

    @property
    def master_pub(self):
        """
        Return the master publish port
        """
        return _get_master_uri(
            self.opts["master_ip"],
            self.publish_port,
            source_ip=self.opts.get("source_ip"),
            source_port=self.opts.get("source_publish_port"),
        )

    @salt.ext.tornado.gen.coroutine
    def _decode_messages(self, messages):
        """
        Take the zmq messages, decrypt/decode them into a payload

        :param list messages: A list of messages to be decoded
        """
        messages_len = len(messages)
        # if it was one message, then its old style
        if messages_len == 1:
            payload = salt.payload.loads(messages[0])
        # 2 includes a header which says who should do it
        elif messages_len == 2:
            message_target = salt.utils.stringutils.to_str(messages[0])
            if (
                self.opts.get("__role") != "syndic"
                and message_target not in ("broadcast", self.hexid)
            ) or (
                self.opts.get("__role") == "syndic"
                and message_target not in ("broadcast", "syndic")
            ):
                log.debug("Publish received for not this minion: %s", message_target)
                raise salt.ext.tornado.gen.Return(None)
            payload = salt.payload.loads(messages[1])
        else:
            raise Exception(
                "Invalid number of messages ({}) in zeromq pubmessage from master".format(
                    len(messages_len)
                )
            )
        # Yield control back to the caller. When the payload has been decoded, assign
        # the decoded payload to 'ret' and resume operation
        ret = yield self._decode_payload(payload)
        raise salt.ext.tornado.gen.Return(ret)

    @property
    def stream(self):
        """
        Return the current zmqstream, creating one if necessary
        """
        if not hasattr(self, "_stream"):
            self._stream = zmq.eventloop.zmqstream.ZMQStream(
                self._socket, io_loop=self.io_loop
            )
        return self._stream

    def on_recv(self, callback):
        """
        Register a callback for received messages (that we didn't initiate)

        :param func callback: A function which should be called when data is received
        """
        if callback is None:
            return self.stream.on_recv(None)

        @salt.ext.tornado.gen.coroutine
        def wrap_callback(messages):
            payload = yield self._decode_messages(messages)
            if payload is not None:
                callback(payload)

        return self.stream.on_recv(wrap_callback)


class ZeroMQReqServerChannel(
    salt.transport.mixins.auth.AESReqServerMixin, salt.transport.server.ReqServerChannel
):
    def __init__(self, opts):
        salt.transport.server.ReqServerChannel.__init__(self, opts)
        self._closing = False
        self._monitor = None
        self._w_monitor = None

    def zmq_device(self):
        """
        Multiprocessing target for the zmq queue device
        """
        self.__setup_signals()
        salt.utils.process.appendproctitle("MWorkerQueue")
        self.context = zmq.Context(self.opts["worker_threads"])
        # Prepare the zeromq sockets
        self.uri = "tcp://{interface}:{ret_port}".format(**self.opts)
        self.clients = self.context.socket(zmq.ROUTER)
        if self.opts["ipv6"] is True and hasattr(zmq, "IPV4ONLY"):
            # IPv6 sockets work for both IPv6 and IPv4 addresses
            self.clients.setsockopt(zmq.IPV4ONLY, 0)
        self.clients.setsockopt(zmq.BACKLOG, self.opts.get("zmq_backlog", 1000))
        self._start_zmq_monitor()
        self.workers = self.context.socket(zmq.DEALER)

        if self.opts["mworker_queue_niceness"] and not salt.utils.platform.is_windows():
            log.info(
                "setting mworker_queue niceness to %d",
                self.opts["mworker_queue_niceness"],
            )
            os.nice(self.opts["mworker_queue_niceness"])

        if self.opts.get("ipc_mode", "") == "tcp":
            self.w_uri = "tcp://127.0.0.1:{}".format(
                self.opts.get("tcp_master_workers", 4515)
            )
        else:
            self.w_uri = "ipc://{}".format(
                os.path.join(self.opts["sock_dir"], "workers.ipc")
            )

        log.info("Setting up the master communication server")
        self.clients.bind(self.uri)
        self.workers.bind(self.w_uri)

        while True:
            if self.clients.closed or self.workers.closed:
                break
            try:
                zmq.device(zmq.QUEUE, self.clients, self.workers)
            except zmq.ZMQError as exc:
                if exc.errno == errno.EINTR:
                    continue
                raise
            except (KeyboardInterrupt, SystemExit):
                break

    def close(self):
        """
        Cleanly shutdown the router socket
        """
        if self._closing:
            return
        log.info("MWorkerQueue under PID %s is closing", os.getpid())
        self._closing = True
        if getattr(self, "_monitor", None) is not None:
            self._monitor.stop()
            self._monitor = None
        if getattr(self, "_w_monitor", None) is not None:
            self._w_monitor.stop()
            self._w_monitor = None
        if hasattr(self, "clients") and self.clients.closed is False:
            self.clients.close()
        if hasattr(self, "workers") and self.workers.closed is False:
            self.workers.close()
        if hasattr(self, "stream"):
            self.stream.close()
        if hasattr(self, "_socket") and self._socket.closed is False:
            self._socket.close()
        if hasattr(self, "context") and self.context.closed is False:
            self.context.term()

    def pre_fork(self, process_manager):
        """
        Pre-fork we need to create the zmq router device

        :param func process_manager: An instance of salt.utils.process.ProcessManager
        """
        salt.transport.mixins.auth.AESReqServerMixin.pre_fork(self, process_manager)
        process_manager.add_process(self.zmq_device)

    def _start_zmq_monitor(self):
        """
        Starts ZMQ monitor for debugging purposes.
        :return:
        """
        # Socket monitor shall be used the only for debug
        # purposes so using threading doesn't look too bad here

        if HAS_ZMQ_MONITOR and self.opts["zmq_monitor"]:
            log.debug("Starting ZMQ monitor")
            import threading

            self._w_monitor = ZeroMQSocketMonitor(self._socket)
            threading.Thread(target=self._w_monitor.start_poll).start()
            log.debug("ZMQ monitor has been started started")

    def post_fork(self, payload_handler, io_loop):
        """
        After forking we need to create all of the local sockets to listen to the
        router

        :param func payload_handler: A function to called to handle incoming payloads as
                                     they are picked up off the wire
        :param IOLoop io_loop: An instance of a Tornado IOLoop, to handle event scheduling
        """
        self.payload_handler = payload_handler
        self.io_loop = io_loop

        self.context = zmq.Context(1)
        self._socket = self.context.socket(zmq.REP)
        self._start_zmq_monitor()

        if self.opts.get("ipc_mode", "") == "tcp":
            self.w_uri = "tcp://127.0.0.1:{}".format(
                self.opts.get("tcp_master_workers", 4515)
            )
        else:
            self.w_uri = "ipc://{}".format(
                os.path.join(self.opts["sock_dir"], "workers.ipc")
            )
        log.info("Worker binding to socket %s", self.w_uri)
        self._socket.connect(self.w_uri)

        salt.transport.mixins.auth.AESReqServerMixin.post_fork(
            self, payload_handler, io_loop
        )

        self.stream = zmq.eventloop.zmqstream.ZMQStream(
            self._socket, io_loop=self.io_loop
        )
        self.stream.on_recv_stream(self.handle_message)

    @salt.ext.tornado.gen.coroutine
    def handle_message(self, stream, payload):
        """
        Handle incoming messages from underlying TCP streams

        :stream ZMQStream stream: A ZeroMQ stream.
        See http://zeromq.github.io/pyzmq/api/generated/zmq.eventloop.zmqstream.html

        :param dict payload: A payload to process
        """
        try:
            payload = salt.payload.loads(payload[0])
            payload = self._decode_payload(payload)
        except Exception as exc:  # pylint: disable=broad-except
            exc_type = type(exc).__name__
            if exc_type == "AuthenticationError":
                log.debug(
                    "Minion failed to auth to master. Since the payload is "
                    "encrypted, it is not known which minion failed to "
                    "authenticate. It is likely that this is a transient "
                    "failure due to the master rotating its public key."
                )
            else:
                log.error("Bad load from minion: %s: %s", exc_type, exc)
            stream.send(salt.payload.dumps("bad load"))
            raise salt.ext.tornado.gen.Return()

        # TODO helper functions to normalize payload?
        if not isinstance(payload, dict) or not isinstance(payload.get("load"), dict):
            log.error(
                "payload and load must be a dict. Payload was: %s and load was %s",
                payload,
                payload.get("load"),
            )
            stream.send(salt.payload.dumps("payload and load must be a dict"))
            raise salt.ext.tornado.gen.Return()

        try:
            id_ = payload["load"].get("id", "")
            if "\0" in id_:
                log.error("Payload contains an id with a null byte: %s", payload)
                stream.send(salt.payload.dumps("bad load: id contains a null byte"))
                raise salt.ext.tornado.gen.Return()
        except TypeError:
            log.error("Payload contains non-string id: %s", payload)
            stream.send(
                salt.payload.dumps("bad load: id {} is not a string".format(id_))
            )
            raise salt.ext.tornado.gen.Return()

        version = 0
        if "version" in payload:
            version = payload["version"]

        sign_messages = False
        if version > 1:
            sign_messages = True

        # intercept the "_auth" commands, since the main daemon shouldn't know
        # anything about our key auth
        if payload["enc"] == "clear" and payload.get("load", {}).get("cmd") == "_auth":
            stream.send(salt.payload.dumps(self._auth(payload["load"], sign_messages)))
            raise salt.ext.tornado.gen.Return()

        nonce = None
        if version > 1:
            nonce = payload["load"].pop("nonce", None)

        # TODO: test
        try:
            # Take the payload_handler function that was registered when we created the channel
            # and call it, returning control to the caller until it completes
            ret, req_opts = yield self.payload_handler(payload)
        except Exception as e:  # pylint: disable=broad-except
            # always attempt to return an error to the minion
            stream.send("Some exception handling minion payload")
            log.error("Some exception handling a payload from minion", exc_info=True)
            raise salt.ext.tornado.gen.Return()

        req_fun = req_opts.get("fun", "send")
        if req_fun == "send_clear":
            stream.send(salt.payload.dumps(ret))
        elif req_fun == "send":
            stream.send(salt.payload.dumps(self.crypticle.dumps(ret, nonce)))
        elif req_fun == "send_private":
            stream.send(
                salt.payload.dumps(
                    self._encrypt_private(
                        ret,
                        req_opts["key"],
                        req_opts["tgt"],
                        nonce,
                        sign_messages,
                    )
                )
            )
        else:
            log.error("Unknown req_fun %s", req_fun)
            # always attempt to return an error to the minion
            stream.send("Server-side exception handling payload")
        raise salt.ext.tornado.gen.Return()

    def __setup_signals(self):
        signal.signal(signal.SIGINT, self._handle_signals)
        signal.signal(signal.SIGTERM, self._handle_signals)

    def _handle_signals(self, signum, sigframe):
        msg = "{} received a ".format(self.__class__.__name__)
        if signum == signal.SIGINT:
            msg += "SIGINT"
        elif signum == signal.SIGTERM:
            msg += "SIGTERM"
        msg += ". Exiting"
        log.debug(msg)
        self.close()
        sys.exit(salt.defaults.exitcodes.EX_OK)


def _set_tcp_keepalive(zmq_socket, opts):
    """
    Ensure that TCP keepalives are set as specified in "opts".

    Warning: Failure to set TCP keepalives on the salt-master can result in
    not detecting the loss of a minion when the connection is lost or when
    its host has been terminated without first closing the socket.
    Salt's Presence System depends on this connection status to know if a minion
    is "present".

    Warning: Failure to set TCP keepalives on minions can result in frequent or
    unexpected disconnects!
    """
    if hasattr(zmq, "TCP_KEEPALIVE") and opts:
        if "tcp_keepalive" in opts:
            zmq_socket.setsockopt(zmq.TCP_KEEPALIVE, opts["tcp_keepalive"])
        if "tcp_keepalive_idle" in opts:
            zmq_socket.setsockopt(zmq.TCP_KEEPALIVE_IDLE, opts["tcp_keepalive_idle"])
        if "tcp_keepalive_cnt" in opts:
            zmq_socket.setsockopt(zmq.TCP_KEEPALIVE_CNT, opts["tcp_keepalive_cnt"])
        if "tcp_keepalive_intvl" in opts:
            zmq_socket.setsockopt(zmq.TCP_KEEPALIVE_INTVL, opts["tcp_keepalive_intvl"])


class ZeroMQPubServerChannel(salt.transport.server.PubServerChannel):
    """
    Encapsulate synchronous operations for a publisher channel
    """

    _sock_data = threading.local()

    def __init__(self, opts):
        self.opts = opts
        self.ckminions = salt.utils.minions.CkMinions(self.opts)

    def connect(self):
        return salt.ext.tornado.gen.sleep(5)

    def _publish_daemon(self, log_queue=None):
        """
        Bind to the interface specified in the configuration file
        """
        salt.utils.process.appendproctitle(self.__class__.__name__)

        if self.opts["pub_server_niceness"] and not salt.utils.platform.is_windows():
            log.info(
                "setting Publish daemon niceness to %i",
                self.opts["pub_server_niceness"],
            )
            os.nice(self.opts["pub_server_niceness"])

        if log_queue:
            salt.log.setup.set_multiprocessing_logging_queue(log_queue)
            salt.log.setup.setup_multiprocessing_logging(log_queue)

        # Set up the context
        context = zmq.Context(1)
        # Prepare minion publish socket
        pub_sock = context.socket(zmq.PUB)
        _set_tcp_keepalive(pub_sock, self.opts)
        # if 2.1 >= zmq < 3.0, we only have one HWM setting
        try:
            pub_sock.setsockopt(zmq.HWM, self.opts.get("pub_hwm", 1000))
        # in zmq >= 3.0, there are separate send and receive HWM settings
        except AttributeError:
            # Set the High Water Marks. For more information on HWM, see:
            # http://api.zeromq.org/4-1:zmq-setsockopt
            pub_sock.setsockopt(zmq.SNDHWM, self.opts.get("pub_hwm", 1000))
            pub_sock.setsockopt(zmq.RCVHWM, self.opts.get("pub_hwm", 1000))
        if self.opts["ipv6"] is True and hasattr(zmq, "IPV4ONLY"):
            # IPv6 sockets work for both IPv6 and IPv4 addresses
            pub_sock.setsockopt(zmq.IPV4ONLY, 0)
        pub_sock.setsockopt(zmq.BACKLOG, self.opts.get("zmq_backlog", 1000))
        pub_sock.setsockopt(zmq.LINGER, -1)
        pub_uri = "tcp://{interface}:{publish_port}".format(**self.opts)
        # Prepare minion pull socket
        pull_sock = context.socket(zmq.PULL)
        pull_sock.setsockopt(zmq.LINGER, -1)

        if self.opts.get("ipc_mode", "") == "tcp":
            pull_uri = "tcp://127.0.0.1:{}".format(
                self.opts.get("tcp_master_publish_pull", 4514)
            )
        else:
            pull_uri = "ipc://{}".format(
                os.path.join(self.opts["sock_dir"], "publish_pull.ipc")
            )
        salt.utils.zeromq.check_ipc_path_max_len(pull_uri)

        # Start the minion command publisher
        log.info("Starting the Salt Publisher on %s", pub_uri)
        pub_sock.bind(pub_uri)

        # Securely create socket
        log.info("Starting the Salt Puller on %s", pull_uri)
        with salt.utils.files.set_umask(0o177):
            pull_sock.bind(pull_uri)

        try:
            while True:
                # Catch and handle EINTR from when this process is sent
                # SIGUSR1 gracefully so we don't choke and die horribly
                try:
                    log.debug("Publish daemon getting data from puller %s", pull_uri)
                    package = pull_sock.recv()
                    package = salt.payload.loads(package)
                    package = self.pack_publish(package)
                    log.debug("Publish daemon received payload. size=%d", len(package))

                    unpacked_package = salt.payload.unpackage(package)
                    unpacked_package = salt.transport.frame.decode_embedded_strs(
                        unpacked_package
                    )
                    payload = unpacked_package["payload"]
                    log.trace("Accepted unpacked package from puller")
                    if self.opts["zmq_filtering"]:
                        # if you have a specific topic list, use that
                        if "topic_lst" in unpacked_package:
                            for topic in unpacked_package["topic_lst"]:
                                log.trace(
                                    "Sending filtered data over publisher %s", pub_uri
                                )
                                # zmq filters are substring match, hash the topic
                                # to avoid collisions
                                htopic = salt.utils.stringutils.to_bytes(
                                    hashlib.sha1(
                                        salt.utils.stringutils.to_bytes(topic)
                                    ).hexdigest()
                                )
                                pub_sock.send(htopic, flags=zmq.SNDMORE)
                                pub_sock.send(payload)
                                log.trace("Filtered data has been sent")

                            # Syndic broadcast
                            if self.opts.get("order_masters"):
                                log.trace("Sending filtered data to syndic")
                                pub_sock.send(b"syndic", flags=zmq.SNDMORE)
                                pub_sock.send(payload)
                                log.trace("Filtered data has been sent to syndic")
                        # otherwise its a broadcast
                        else:
                            # TODO: constants file for "broadcast"
                            log.trace(
                                "Sending broadcasted data over publisher %s", pub_uri
                            )
                            pub_sock.send(b"broadcast", flags=zmq.SNDMORE)
                            pub_sock.send(payload)
                            log.trace("Broadcasted data has been sent")
                    else:
                        log.trace(
                            "Sending ZMQ-unfiltered data over publisher %s", pub_uri
                        )
                        pub_sock.send(payload)
                        log.trace("Unfiltered data has been sent")
                except zmq.ZMQError as exc:
                    if exc.errno == errno.EINTR:
                        continue
                    raise

        except KeyboardInterrupt:
            log.trace("Publish daemon caught Keyboard interupt, tearing down")
        # Cleanly close the sockets if we're shutting down
        if pub_sock.closed is False:
            pub_sock.close()
        if pull_sock.closed is False:
            pull_sock.close()
        if context.closed is False:
            context.term()

    def pre_fork(self, process_manager, kwargs=None):
        """
        Do anything necessary pre-fork. Since this is on the master side this will
        primarily be used to create IPC channels and create our daemon process to
        do the actual publishing

        :param func process_manager: A ProcessManager, from salt.utils.process.ProcessManager
        """
        process_manager.add_process(self._publish_daemon, kwargs=kwargs)

    @property
    def pub_sock(self):
        """
        This thread's zmq publisher socket. This socket is stored on the class
        so that multiple instantiations in the same thread will re-use a single
        zmq socket.
        """
        try:
            return self._sock_data.sock
        except AttributeError:
            pass

    def pub_connect(self):
        """
        Create and connect this thread's zmq socket. If a publisher socket
        already exists "pub_close" is called before creating and connecting a
        new socket.
        """
        if self.pub_sock:
            self.pub_close()
        self._sock_data._ctx = zmq.Context()
        self._sock_data.sock = self._sock_data._ctx.socket(zmq.PUSH)
        self.pub_sock.setsockopt(zmq.LINGER, -1)
        if self.opts.get("ipc_mode", "") == "tcp":
            pull_uri = "tcp://127.0.0.1:{}".format(
                self.opts.get("tcp_master_publish_pull", 4514)
            )
        else:
            pull_uri = "ipc://{}".format(
                os.path.join(self.opts["sock_dir"], "publish_pull.ipc")
            )
        log.debug("Connecting to pub server: %s", pull_uri)
        self.pub_sock.connect(pull_uri)
        return self._sock_data.sock

    def pub_close(self):
        """
        Disconnect an existing publisher socket and remove it from the local
        thread's cache.
        """
        if hasattr(self._sock_data, "sock"):
            self._sock_data.sock.close()
            delattr(self._sock_data, "sock")
            if hasattr(self._sock_data, "_ctx"):
                self._sock_data._ctx.destroy()

    def pack_publish(self, load):
        payload = {"enc": "aes"}
        load["serial"] = salt.master.SMaster.get_serial()
        crypticle = salt.crypt.Crypticle(
            self.opts, salt.master.SMaster.secrets["aes"]["secret"].value
        )
        payload["load"] = crypticle.dumps(load)
        if self.opts["sign_pub_messages"]:
            master_pem_path = os.path.join(self.opts["pki_dir"], "master.pem")
            log.debug("Signing data packet")
            payload["sig"] = salt.crypt.sign_message(master_pem_path, payload["load"])
        int_payload = {"payload": salt.payload.dumps(payload)}

        # add some targeting stuff for lists only (for now)
        if load["tgt_type"] == "list":
            int_payload["topic_lst"] = load["tgt"]

        # If zmq_filtering is enabled, target matching has to happen master side
        match_targets = ["pcre", "glob", "list"]
        if self.opts["zmq_filtering"] and load["tgt_type"] in match_targets:
            # Fetch a list of minions that match
            _res = self.ckminions.check_minions(load["tgt"], tgt_type=load["tgt_type"])
            match_ids = _res["minions"]

            log.debug("Publish Side Match: %s", match_ids)
            # Send list of miions thru so zmq can target them
            int_payload["topic_lst"] = match_ids
        payload = salt.payload.dumps(int_payload)
        log.debug(
            "Sending payload to publish daemon. jid=%s size=%d",
            load.get("jid", None),
            len(payload),
        )
        return payload

    def publish(self, load):
        """
        Publish "load" to minions. This send the load to the publisher daemon
        process with does the actual sending to minions.

        :param dict load: A load to be sent across the wire to minions
        """
        if not self.pub_sock:
            self.pub_connect()
        self.pub_sock.send(salt.payload.dumps(load))
        log.debug("Sent payload to publish daemon.")


class AsyncReqMessageClientPool(salt.transport.MessageClientPool):
    """
    Wrapper class of AsyncReqMessageClientPool to avoid blocking waiting while writing data to socket.
    """

    def __init__(self, opts, args=None, kwargs=None):
        self._closing = False
        super().__init__(AsyncReqMessageClient, opts, args=args, kwargs=kwargs)

    def close(self):
        if self._closing:
            return

        self._closing = True
        for message_client in self.message_clients:
            message_client.close()
        self.message_clients = []

    def send(self, *args, **kwargs):
        message_clients = sorted(self.message_clients, key=lambda x: len(x.send_queue))
        return message_clients[0].send(*args, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()


# TODO: unit tests!
class AsyncReqMessageClient:
    """
    This class wraps the underlying zeromq REQ socket and gives a future-based
    interface to sending and recieving messages. This works around the primary
    limitation of serialized send/recv on the underlying socket by queueing the
    message sends in this class. In the future if we decide to attempt to multiplex
    we can manage a pool of REQ/REP sockets-- but for now we'll just do them in serial
    """

    def __init__(self, opts, addr, linger=0, io_loop=None):
        """
        Create an asynchronous message client

        :param dict opts: The salt opts dictionary
        :param str addr: The interface IP address to bind to
        :param int linger: The number of seconds to linger on a ZMQ socket. See
                           http://api.zeromq.org/2-1:zmq-setsockopt [ZMQ_LINGER]
        :param IOLoop io_loop: A Tornado IOLoop event scheduler [tornado.ioloop.IOLoop]
        """
        self.opts = opts
        self.addr = addr
        self.linger = linger
        if io_loop is None:
            self.io_loop = salt.ext.tornado.ioloop.IOLoop.current()
        else:
            self.io_loop = io_loop

        self.context = zmq.Context()

        # wire up sockets
        self._init_socket()

        self.send_queue = []
        # mapping of message -> future
        self.send_future_map = {}

        self.send_timeout_map = {}  # message -> timeout
        self._closing = False

    # TODO: timeout all in-flight sessions, or error
    def close(self):
        try:
            if self._closing:
                return
        except AttributeError:
            # We must have been called from __del__
            # The python interpreter has nuked most attributes already
            return
        else:
            self._closing = True
            if hasattr(self, "stream") and self.stream is not None:
                if ZMQ_VERSION_INFO < (14, 3, 0):
                    # stream.close() doesn't work properly on pyzmq < 14.3.0
                    if self.stream.socket:
                        self.stream.socket.close()
                    self.stream.io_loop.remove_handler(self.stream.socket)
                    # set this to None, more hacks for messed up pyzmq
                    self.stream.socket = None
                    self.socket.close()
                else:
                    self.stream.close()
                    self.socket = None
                self.stream = None
            if self.context.closed is False:
                self.context.term()

    # pylint: disable=W1701
    def __del__(self):
        self.close()

    # pylint: enable=W1701

    def _init_socket(self):
        if hasattr(self, "stream"):
            self.stream.close()  # pylint: disable=E0203
            self.socket.close()  # pylint: disable=E0203
            del self.stream
            del self.socket

        self.socket = self.context.socket(zmq.REQ)

        # socket options
        if hasattr(zmq, "RECONNECT_IVL_MAX"):
            self.socket.setsockopt(zmq.RECONNECT_IVL_MAX, 5000)

        _set_tcp_keepalive(self.socket, self.opts)
        if self.addr.startswith("tcp://["):
            # Hint PF type if bracket enclosed IPv6 address
            if hasattr(zmq, "IPV6"):
                self.socket.setsockopt(zmq.IPV6, 1)
            elif hasattr(zmq, "IPV4ONLY"):
                self.socket.setsockopt(zmq.IPV4ONLY, 0)
        self.socket.linger = self.linger
        log.debug("Trying to connect to: %s", self.addr)
        self.socket.connect(self.addr)
        self.stream = zmq.eventloop.zmqstream.ZMQStream(
            self.socket, io_loop=self.io_loop
        )

    @salt.ext.tornado.gen.coroutine
    def _internal_send_recv(self):
        while len(self.send_queue) > 0:
            message = self.send_queue[0]
            future = self.send_future_map.get(message, None)
            if future is None:
                # Timedout
                del self.send_queue[0]
                continue

            # send
            def mark_future(msg):
                if not future.done():
                    data = salt.payload.loads(msg[0])
                    future.set_result(data)

            self.stream.on_recv(mark_future)
            self.stream.send(message)

            try:
                ret = yield future
            except Exception as err:  # pylint: disable=broad-except
                log.debug("Re-init ZMQ socket: %s", err)
                self._init_socket()  # re-init the zmq socket (no other way in zmq)
                del self.send_queue[0]
                continue
            del self.send_queue[0]
            self.send_future_map.pop(message, None)
            self.remove_message_timeout(message)

    def remove_message_timeout(self, message):
        if message not in self.send_timeout_map:
            return
        timeout = self.send_timeout_map.pop(message, None)
        if timeout is not None:
            # Hasn't been already timedout
            self.io_loop.remove_timeout(timeout)

    def timeout_message(self, message):
        """
        Handle a message timeout by removing it from the sending queue
        and informing the caller

        :raises: SaltReqTimeoutError
        """
        future = self.send_future_map.pop(message, None)
        # In a race condition the message might have been sent by the time
        # we're timing it out. Make sure the future is not None
        if future is not None:
            del self.send_timeout_map[message]
            if future.attempts < future.tries:
                future.attempts += 1
                log.debug(
                    "SaltReqTimeoutError, retrying. (%s/%s)",
                    future.attempts,
                    future.tries,
                )
                self.send(
                    message,
                    timeout=future.timeout,
                    tries=future.tries,
                    future=future,
                )

            else:
                future.set_exception(SaltReqTimeoutError("Message timed out"))

    def send(
        self, message, timeout=None, tries=3, future=None, callback=None, raw=False
    ):
        """
        Return a future which will be completed when the message has a response
        """
        if future is None:
            future = salt.ext.tornado.concurrent.Future()
            future.tries = tries
            future.attempts = 0
            future.timeout = timeout
            # if a future wasn't passed in, we need to serialize the message
            message = salt.payload.dumps(message)
        if callback is not None:

            def handle_future(future):
                response = future.result()
                self.io_loop.add_callback(callback, response)

            future.add_done_callback(handle_future)
        # Add this future to the mapping
        self.send_future_map[message] = future

        if self.opts.get("detect_mode") is True:
            timeout = 1

        if timeout is not None:
            send_timeout = self.io_loop.call_later(
                timeout, self.timeout_message, message
            )
            self.send_timeout_map[message] = send_timeout

        if len(self.send_queue) == 0:
            self.io_loop.spawn_callback(self._internal_send_recv)

        self.send_queue.append(message)

        return future


class ZeroMQSocketMonitor:
    __EVENT_MAP = None

    def __init__(self, socket):
        """
        Create ZMQ monitor sockets

        More information:
            http://api.zeromq.org/4-0:zmq-socket-monitor
        """
        self._socket = socket
        self._monitor_socket = self._socket.get_monitor_socket()
        self._monitor_stream = None

    def start_io_loop(self, io_loop):
        log.trace("Event monitor start!")
        self._monitor_stream = zmq.eventloop.zmqstream.ZMQStream(
            self._monitor_socket, io_loop=io_loop
        )
        self._monitor_stream.on_recv(self.monitor_callback)

    def start_poll(self):
        log.trace("Event monitor start!")
        try:
            while self._monitor_socket is not None and self._monitor_socket.poll():
                msg = self._monitor_socket.recv_multipart()
                self.monitor_callback(msg)
        except (AttributeError, zmq.error.ContextTerminated):
            # We cannot log here because we'll get an interrupted system call in trying
            # to flush the logging buffer as we terminate
            pass

    @property
    def event_map(self):
        if ZeroMQSocketMonitor.__EVENT_MAP is None:
            event_map = {}
            for name in dir(zmq):
                if name.startswith("EVENT_"):
                    value = getattr(zmq, name)
                    event_map[value] = name
            ZeroMQSocketMonitor.__EVENT_MAP = event_map
        return ZeroMQSocketMonitor.__EVENT_MAP

    def monitor_callback(self, msg):
        evt = zmq.utils.monitor.parse_monitor_message(msg)
        evt["description"] = self.event_map[evt["event"]]
        log.debug("ZeroMQ event: %s", evt)
        if evt["event"] == zmq.EVENT_MONITOR_STOPPED:
            self.stop()

    def stop(self):
        if self._socket is None:
            return
        self._socket.disable_monitor()
        self._socket = None
        self._monitor_socket = None
        if self._monitor_stream is not None:
            self._monitor_stream.close()
            self._monitor_stream = None
        log.trace("Event monitor done!")



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest
import salt.states.supervisord as supervisord
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {supervisord: {}}


def test_running():
    """
    Test to ensure the named service is running.
    """
    name = "wsgi_server"

    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    comt = "Supervisord module not activated. Do you need to install supervisord?"
    ret.update({"comment": comt, "result": False})
    assert supervisord.running(name) == ret

    mock = MagicMock(return_value={name: {"state": "running"}})
    with patch.dict(supervisord.__salt__, {"supervisord.status": mock}):
        with patch.dict(supervisord.__opts__, {"test": True}):
            comt = "Service wsgi_server is already running"
            ret.update({"comment": comt, "result": True})
            assert supervisord.running(name) == ret

        with patch.dict(supervisord.__opts__, {"test": False}):
            comt = "Not starting already running service: wsgi_server"
            ret.update({"comment": comt})
            assert supervisord.running(name) == ret


def test_dead():
    """
    Test to ensure the named service is dead (not running).
    """
    name = "wsgi_server"

    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    with patch.dict(supervisord.__opts__, {"test": True}):
        comt = "Service {} is set to be stopped".format(name)
        ret.update({"comment": comt})
        assert supervisord.dead(name) == ret


def test_mod_watch():
    """
    Test to always restart on watch.
    """
    name = "wsgi_server"

    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    comt = "Supervisord module not activated. Do you need to install supervisord?"
    ret.update({"comment": comt, "result": False})
    assert supervisord.mod_watch(name) == ret



import xml.etree.ElementTree as ET

import pytest
import salt.utils.xmlutil as xml


@pytest.fixture
def xml_doc():
    return ET.fromstring(
        """
        <domain>
            <name>test01</name>
            <memory unit="MiB">1024</memory>
            <cpu>
                <topology sockets="1"/>
            </cpu>
            <vcpus>
              <vcpu enabled="yes" id="1"/>
            </vcpus>
            <memtune>
              <hugepages>
                <page size="128"/>
              </hugepages>
            </memtune>
        </domain>
    """
    )


def test_change_xml_text(xml_doc):
    ret = xml.change_xml(
        xml_doc, {"name": "test02"}, [{"path": "name", "xpath": "name"}]
    )
    assert ret
    assert "test02" == xml_doc.find("name").text


def test_change_xml_text_nochange(xml_doc):
    ret = xml.change_xml(
        xml_doc, {"name": "test01"}, [{"path": "name", "xpath": "name"}]
    )
    assert not ret


def test_change_xml_equals_nochange(xml_doc):
    ret = xml.change_xml(
        xml_doc,
        {"mem": 1023},
        [
            {
                "path": "mem",
                "xpath": "memory",
                "get": lambda n: int(n.text),
                "equals": lambda o, n: abs(o - n) <= 1,
            }
        ],
    )
    assert not ret


def test_change_xml_text_notdefined(xml_doc):
    ret = xml.change_xml(xml_doc, {}, [{"path": "name", "xpath": "name"}])
    assert not ret


def test_change_xml_text_removed(xml_doc):
    ret = xml.change_xml(xml_doc, {"name": None}, [{"path": "name", "xpath": "name"}])
    assert ret
    assert xml_doc.find("name") is None


def test_change_xml_text_add(xml_doc):
    ret = xml.change_xml(
        xml_doc,
        {"cpu": {"vendor": "ACME"}},
        [{"path": "cpu:vendor", "xpath": "cpu/vendor"}],
    )
    assert ret
    assert "ACME" == xml_doc.find("cpu/vendor").text


def test_change_xml_convert(xml_doc):
    ret = xml.change_xml(
        xml_doc,
        {"mem": 2},
        [{"path": "mem", "xpath": "memory", "convert": lambda v: v * 1024}],
    )
    assert ret
    assert "2048" == xml_doc.find("memory").text


def test_change_xml_attr(xml_doc):
    ret = xml.change_xml(
        xml_doc,
        {"cpu": {"topology": {"cores": 4}}},
        [
            {
                "path": "cpu:topology:cores",
                "xpath": "cpu/topology",
                "get": lambda n: int(n.get("cores")) if n.get("cores") else None,
                "set": lambda n, v: n.set("cores", str(v)),
                "del": xml.del_attribute("cores"),
            }
        ],
    )
    assert ret
    assert "4" == xml_doc.find("cpu/topology").get("cores")


def test_change_xml_attr_unchanged(xml_doc):
    ret = xml.change_xml(
        xml_doc,
        {"cpu": {"topology": {"sockets": 1}}},
        [
            {
                "path": "cpu:topology:sockets",
                "xpath": "cpu/topology",
                "get": lambda n: int(n.get("sockets")) if n.get("sockets") else None,
                "set": lambda n, v: n.set("sockets", str(v)),
                "del": xml.del_attribute("sockets"),
            }
        ],
    )
    assert not ret


def test_change_xml_attr_remove(xml_doc):
    ret = xml.change_xml(
        xml_doc,
        {"cpu": {"topology": {"sockets": None}}},
        [
            {
                "path": "cpu:topology:sockets",
                "xpath": "./cpu/topology",
                "get": lambda n: int(n.get("sockets")) if n.get("sockets") else None,
                "set": lambda n, v: n.set("sockets", str(v)),
                "del": xml.del_attribute("sockets"),
            }
        ],
    )
    assert ret
    assert xml_doc.find("cpu") is None


def test_change_xml_not_simple_value(xml_doc):
    ret = xml.change_xml(
        xml_doc,
        {"cpu": {"topology": {"sockets": None}}},
        [{"path": "cpu", "xpath": "vcpu", "get": lambda n: int(n.text)}],
    )
    assert not ret


def test_change_xml_template(xml_doc):
    ret = xml.change_xml(
        xml_doc,
        {"cpu": {"vcpus": {2: {"enabled": True}, 4: {"enabled": False}}}},
        [
            {
                "path": "cpu:vcpus:{id}:enabled",
                "xpath": "vcpus/vcpu[@id='$id']",
                "convert": lambda v: "yes" if v else "no",
                "get": lambda n: n.get("enabled"),
                "set": lambda n, v: n.set("enabled", v),
                "del": xml.del_attribute("enabled", ["id"]),
            },
        ],
    )
    assert ret
    assert xml_doc.find("vcpus/vcpu[@id='1']") is None
    assert "yes" == xml_doc.find("vcpus/vcpu[@id='2']").get("enabled")
    assert "no" == xml_doc.find("vcpus/vcpu[@id='4']").get("enabled")


def test_change_xml_template_remove(xml_doc):
    ret = xml.change_xml(
        xml_doc,
        {"cpu": {"vcpus": None}},
        [
            {
                "path": "cpu:vcpus:{id}:enabled",
                "xpath": "vcpus/vcpu[@id='$id']",
                "convert": lambda v: "yes" if v else "no",
                "get": lambda n: n.get("enabled"),
                "set": lambda n, v: n.set("enabled", v),
                "del": xml.del_attribute("enabled", ["id"]),
            },
        ],
    )
    assert ret
    assert xml_doc.find("vcpus") is None


def test_change_xml_template_list(xml_doc):
    ret = xml.change_xml(
        xml_doc,
        {"memtune": {"hugepages": [{"size": "1024"}, {"size": "512"}]}},
        [
            {
                "path": "memtune:hugepages:{id}:size",
                "xpath": "memtune/hugepages/page[$id]",
                "get": lambda n: n.get("size"),
                "set": lambda n, v: n.set("size", v),
                "del": xml.del_attribute("size"),
            },
        ],
    )
    assert ret
    assert ["1024", "512"] == [
        n.get("size") for n in xml_doc.findall("memtune/hugepages/page")
    ]


def test_strip_spaces():
    xml_str = """<domain>
            <name>test01</name>
            <memory unit="MiB" >1024</memory> 
        </domain>
    """
    expected_str = (
        b'<domain><name>test01</name><memory unit="MiB">1024</memory></domain>'
    )

    node = ET.fromstring(xml_str)
    assert expected_str == ET.tostring(xml.strip_spaces(node))



"""
An engine that reads messages from the redis sentinel pubsub and sends reactor
events based on the channels they are subscribed to.

.. versionadded:: 2016.3.0

:configuration:

    Example configuration

    .. code-block:: yaml

        engines:
          - redis_sentinel:
              hosts:
                matching: 'board*'
                port: 26379
                interface: eth2
              channels:
                - '+switch-master'
                - '+odown'
                - '-odown'

:depends: redis
"""

import logging

import salt.client

try:
    import redis
except ImportError:
    redis = None

log = logging.getLogger(__name__)

__virtualname__ = "redis"


def __virtual__():
    return (
        __virtualname__
        if redis is not None
        else (False, "redis python module is not installed")
    )


class Listener:
    def __init__(self, host=None, port=None, channels=None, tag=None):
        if host is None:
            host = "localhost"
        if port is None:
            port = 26379
        if channels is None:
            channels = ["*"]
        if tag is None:
            tag = "salt/engine/redis_sentinel"
        super().__init__()
        self.tag = tag
        self.redis = redis.StrictRedis(host=host, port=port, decode_responses=True)
        self.pubsub = self.redis.pubsub()
        self.pubsub.psubscribe(channels)
        self.fire_master = salt.utils.event.get_master_event(
            __opts__, __opts__["sock_dir"]
        ).fire_event

    def work(self, item):
        ret = {"channel": item["channel"]}
        if isinstance(item["data"], int):
            ret["code"] = item["data"]
        elif item["channel"] == "+switch-master":
            ret.update(
                dict(
                    list(
                        zip(
                            ("master", "old_host", "old_port", "new_host", "new_port"),
                            item["data"].split(" "),
                        )
                    )
                )
            )
        elif item["channel"] in ("+odown", "-odown"):
            ret.update(
                dict(list(zip(("master", "host", "port"), item["data"].split(" ")[1:])))
            )
        else:
            ret = {
                "channel": item["channel"],
                "data": item["data"],
            }
        self.fire_master(ret, "{}/{}".format(self.tag, item["channel"]))

    def run(self):
        log.debug("Start Listener")
        for item in self.pubsub.listen():
            log.debug("Item: %s", item)
            self.work(item)


def start(hosts, channels, tag=None):
    if tag is None:
        tag = "salt/engine/redis_sentinel"
    with salt.client.LocalClient() as local:
        ips = local.cmd(
            hosts["matching"], "network.ip_addrs", [hosts["interface"]]
        ).values()
    client = Listener(host=ips.pop()[0], port=hosts["port"], channels=channels, tag=tag)
    client.run()



import logging

import salt.config
from tests.support.helpers import PRE_PYTEST_SKIP

# Using the PRE_PYTEST_SKIP decorator since this test still fails on some platforms.
# Will investigate later.


@PRE_PYTEST_SKIP
def test_jid_in_logs(caplog, salt_call_cli):
    """
    Test JID in log_format
    """
    jid_formatted_str = salt.config._DFLT_LOG_FMT_JID.split("%")[0]
    formatter = logging.Formatter(fmt="%(jid)s %(message)s")
    with caplog.at_level(logging.DEBUG):
        previous_formatter = caplog.handler.formatter
        try:
            caplog.handler.setFormatter(formatter)
            ret = salt_call_cli.run("test.ping")
            assert ret.exitcode == 0
            assert ret.json is True

            assert_error_msg = (
                "'{}' not found in log messages:\n>>>>>>>>>{}\n<<<<<<<<<".format(
                    jid_formatted_str, caplog.text
                )
            )
            assert jid_formatted_str in caplog.text, assert_error_msg
        finally:
            caplog.handler.setFormatter(previous_formatter)



"""
    :codeauthor: `Gareth J. Greenaway <gareth@saltstack.com>`

    tests.unit.cloud.clouds.digitalocean_test
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""


import logging

from salt.cloud.clouds import digitalocean
from salt.exceptions import SaltCloudSystemExit
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class DigitalOceanTestCase(TestCase):
    """
    Unit TestCase for salt.cloud.clouds.digitalocean module.
    """

    def test_reboot_no_call(self):
        """
        Tests that a SaltCloudSystemExit is raised when
        kwargs that are provided do not include an action.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            digitalocean.reboot,
            name="fake_name",
        )

        with self.assertRaises(SaltCloudSystemExit) as excinfo:
            ret = digitalocean.reboot(name="fake_name")
        self.assertEqual(
            "The reboot action must be called with -a or --action.",
            excinfo.exception.strerror,
        )



import copy

import salt.utils.configcomparer as configcomparer
from tests.support.unit import TestCase


class UtilConfigcomparerTestCase(TestCase):

    base_config = {
        "attr1": "value1",
        "attr2": ["item1", "item2", "item3"],
        "attr3": [],
        "attr4": {},
        "attr5": {"subattr1": "value1", "subattr2": ["item1"]},
    }

    def test_compare_and_update_config(self):

        # empty config
        to_update = copy.deepcopy(self.base_config)
        changes = {}
        configcomparer.compare_and_update_config(
            {},
            to_update,
            changes,
        )
        self.assertEqual({}, changes)
        self.assertEqual(self.base_config, to_update)

        # simple, new value
        to_update = copy.deepcopy(self.base_config)
        changes = {}
        configcomparer.compare_and_update_config(
            {"attrx": "value1"},
            to_update,
            changes,
        )
        self.assertEqual(
            {"attrx": {"new": "value1", "old": None}},
            changes,
        )
        self.assertEqual("value1", to_update["attrx"])
        self.assertEqual("value1", to_update["attr1"])
        # simple value
        to_update = copy.deepcopy(self.base_config)
        changes = {}
        configcomparer.compare_and_update_config(
            {"attr1": "value2"},
            to_update,
            changes,
        )
        self.assertEqual(
            {"attr1": {"new": "value2", "old": "value1"}},
            changes,
        )
        self.assertEqual("value2", to_update["attr1"])
        self.assertEqual(
            {
                "attr1": "value2",
                "attr2": ["item1", "item2", "item3"],
                "attr3": [],
                "attr4": {},
                "attr5": {"subattr1": "value1", "subattr2": ["item1"]},
            },
            to_update,
        )

        # empty list
        to_update = copy.deepcopy(self.base_config)
        changes = {}
        configcomparer.compare_and_update_config(
            {"attr3": []},
            to_update,
            changes,
        )
        self.assertEqual({}, changes)
        self.assertEqual(self.base_config, to_update)

        # list value (add)
        to_update = copy.deepcopy(self.base_config)
        changes = {}
        configcomparer.compare_and_update_config(
            {"attr2": ["item1", "item2", "item3", "item4"]},
            to_update,
            changes,
        )
        self.assertEqual(
            {"attr2[3]": {"new": "item4", "old": None}},
            changes,
        )
        self.assertEqual(
            {
                "attr1": "value1",
                "attr2": ["item1", "item2", "item3", "item4"],
                "attr3": [],
                "attr4": {},
                "attr5": {"subattr1": "value1", "subattr2": ["item1"]},
            },
            to_update,
        )

        # list value (remove and modify)
        to_update = copy.deepcopy(self.base_config)
        changes = {}
        configcomparer.compare_and_update_config(
            {"attr2": ["itemx", "item2"]},
            to_update,
            changes,
        )
        self.assertEqual(
            {
                "attr2[0]": {"new": "itemx", "old": "item1"},
                "attr2[2]": {"new": None, "old": "item3"},
            },
            changes,
        )
        self.assertEqual(
            {
                "attr1": "value1",
                "attr2": ["itemx", "item2"],
                "attr3": [],
                "attr4": {},
                "attr5": {"subattr1": "value1", "subattr2": ["item1"]},
            },
            to_update,
        )

        # empty dict
        to_update = copy.deepcopy(self.base_config)
        changes = {}
        configcomparer.compare_and_update_config(
            {"attr4": {}},
            to_update,
            changes,
        )
        self.assertEqual({}, changes)
        self.assertEqual(self.base_config, to_update)

        # dict value (add)
        to_update = copy.deepcopy(self.base_config)
        changes = {}
        configcomparer.compare_and_update_config(
            {"attr5": {"subattr3": "value1"}},
            to_update,
            changes,
        )
        self.assertEqual(
            {"attr5.subattr3": {"new": "value1", "old": None}},
            changes,
        )
        self.assertEqual(
            {
                "attr1": "value1",
                "attr2": ["item1", "item2", "item3"],
                "attr3": [],
                "attr4": {},
                "attr5": {
                    "subattr1": "value1",
                    "subattr2": ["item1"],
                    "subattr3": "value1",
                },
            },
            to_update,
        )

        # dict value (remove and modify)
        to_update = copy.deepcopy(self.base_config)
        changes = {}
        configcomparer.compare_and_update_config(
            {"attr5": {"subattr1": "value2", "subattr2": ["item1", "item2"]}},
            to_update,
            changes,
        )
        self.assertEqual(
            {
                "attr5.subattr1": {"new": "value2", "old": "value1"},
                "attr5.subattr2[1]": {"new": "item2", "old": None},
            },
            changes,
        )
        self.assertEqual(
            {
                "attr1": "value1",
                "attr2": ["item1", "item2", "item3"],
                "attr3": [],
                "attr4": {},
                "attr5": {"subattr1": "value2", "subattr2": ["item1", "item2"]},
            },
            to_update,
        )



"""
Unit Tests for the mac_desktop execution module.
"""

import salt.modules.mac_desktop as mac_desktop
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class MacDesktopTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.mac_desktop
    """

    def setup_loader_modules(self):
        return {mac_desktop: {}}

    # 'get_output_volume' function tests: 2

    def test_get_output_volume(self):
        """
        Test if it get the output volume (range 0 to 100)
        """
        mock = MagicMock(return_value={"retcode": 0, "stdout": "25"})
        with patch.dict(mac_desktop.__salt__, {"cmd.run_all": mock}):
            self.assertEqual(mac_desktop.get_output_volume(), "25")

    def test_get_output_volume_error(self):
        """
        Tests that an error is raised when cmd.run_all errors
        """
        mock = MagicMock(return_value={"retcode": 1})
        with patch.dict(mac_desktop.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(CommandExecutionError, mac_desktop.get_output_volume)

    # 'set_output_volume' function tests: 2

    def test_set_output_volume(self):
        """
        Test if it set the volume of sound (range 0 to 100)
        """
        mock = MagicMock(return_value={"retcode": 0})
        with patch.dict(mac_desktop.__salt__, {"cmd.run_all": mock}), patch(
            "salt.modules.mac_desktop.get_output_volume", MagicMock(return_value="25")
        ):
            self.assertTrue(mac_desktop.set_output_volume("25"))

    def test_set_output_volume_error(self):
        """
        Tests that an error is raised when cmd.run_all errors
        """
        mock = MagicMock(return_value={"retcode": 1})
        with patch.dict(mac_desktop.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(
                CommandExecutionError, mac_desktop.set_output_volume, "25"
            )

    # 'screensaver' function tests: 2

    def test_screensaver(self):
        """
        Test if it launch the screensaver
        """
        mock = MagicMock(return_value={"retcode": 0})
        with patch.dict(mac_desktop.__salt__, {"cmd.run_all": mock}):
            self.assertTrue(mac_desktop.screensaver())

    def test_screensaver_error(self):
        """
        Tests that an error is raised when cmd.run_all errors
        """
        mock = MagicMock(return_value={"retcode": 1})
        with patch.dict(mac_desktop.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(CommandExecutionError, mac_desktop.screensaver)

    # 'lock' function tests: 2

    def test_lock(self):
        """
        Test if it lock the desktop session
        """
        mock = MagicMock(return_value={"retcode": 0})
        with patch.dict(mac_desktop.__salt__, {"cmd.run_all": mock}):
            self.assertTrue(mac_desktop.lock())

    def test_lock_error(self):
        """
        Tests that an error is raised when cmd.run_all errors
        """
        mock = MagicMock(return_value={"retcode": 1})
        with patch.dict(mac_desktop.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(CommandExecutionError, mac_desktop.lock)

    # 'say' function tests: 2

    def test_say(self):
        """
        Test if it says some words.
        """
        mock = MagicMock(return_value={"retcode": 0})
        with patch.dict(mac_desktop.__salt__, {"cmd.run_all": mock}):
            self.assertTrue(mac_desktop.say())

    def test_say_error(self):
        """
        Tests that an error is raised when cmd.run_all errors
        """
        mock = MagicMock(return_value={"retcode": 1})
        with patch.dict(mac_desktop.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(CommandExecutionError, mac_desktop.say)



"""
Apache Libcloud Storage State
=============================

Manage cloud storage using libcloud

:codeauthor: ``Anthony Shaw <anthonyshaw@apache.org>``

Apache Libcloud Storage (object/blob) management for a full list
of supported clouds, see http://libcloud.readthedocs.io/en/latest/storage/supported_providers.html

Clouds include Amazon S3, Google Storage, Aliyun, Azure Blobs, Ceph, OpenStack swift

.. versionadded:: 2018.3.0

:configuration:
    This module uses a configuration profile for one or multiple Storage providers

    .. code-block:: yaml

        libcloud_storage:
            profile_test1:
              driver: google_storage
              key: GOOG0123456789ABCXYZ
              secret: mysecret
            profile_test2:
              driver: s3
              key: 12345
              secret: mysecret

Examples
--------

Creating a container and uploading a file
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. code-block:: yaml

    web_things:
      libcloud_storage.container_present:
        name: my_container_name
        profile: profile1
      libcloud_storage.object_present:
        name: my_file.jpg
        container: my_container_name
        path: /path/to/local/file.jpg
        profile: profile1

Downloading a file
~~~~~~~~~~~~~~~~~~

This example will download the file from the remote cloud and keep it locally

.. code-block:: yaml

    web_things:
      libcloud_storage.file_present:
        name: my_file.jpg
        container: my_container_name
        path: /path/to/local/file.jpg
        profile: profile1

:depends: apache-libcloud
"""


import logging

log = logging.getLogger(__name__)


def __virtual__():
    if "libcloud_storage.list_containers" in __salt__:
        return True
    return (False, "libcloud_storage module could not be loaded")


def state_result(result, message, name, changes):
    return {"result": result, "comment": message, "name": name, "changes": changes}


def container_present(name, profile):
    """
    Ensures a container is present.

    :param name: Container name
    :type  name: ``str``

    :param profile: The profile key
    :type  profile: ``str``
    """
    containers = __salt__["libcloud_storage.list_containers"](profile)
    match = [z for z in containers if z["name"] == name]
    if len(match) > 0:
        return state_result(True, "Container already exists", name, {})
    else:
        result = __salt__["libcloud_storage.create_container"](name, profile)
        return state_result(True, "Created new container", name, result)


def container_absent(name, profile):
    """
    Ensures a container is absent.

    :param name: Container name
    :type  name: ``str``

    :param profile: The profile key
    :type  profile: ``str``
    """
    containers = __salt__["libcloud_storage.list_containers"](profile)
    match = [z for z in containers if z["name"] == name]
    if len(match) == 0:
        return state_result(True, "Container already absent", name, {})
    else:
        result = __salt__["libcloud_storage.delete_container"](name, profile)
        return state_result(result, "Deleted container", name, {})


def object_present(container, name, path, profile):
    """
    Ensures a object is presnt.

    :param container: Container name
    :type  container: ``str``

    :param name: Object name in cloud
    :type  name: ``str``

    :param path: Local path to file
    :type  path: ``str``

    :param profile: The profile key
    :type  profile: ``str``
    """
    existing_object = __salt__["libcloud_storage.get_container_object"](
        container, name, profile
    )
    if existing_object is not None:
        return state_result(True, "Object already present", name, {})
    else:
        result = __salt__["libcloud_storage.upload_object"](
            path, container, name, profile
        )
        return state_result(result, "Uploaded object", name, {})


def object_absent(container, name, profile):
    """
    Ensures a object is absent.

    :param container: Container name
    :type  container: ``str``

    :param name: Object name in cloud
    :type  name: ``str``

    :param profile: The profile key
    :type  profile: ``str``
    """
    existing_object = __salt__["libcloud_storage.get_container_object"](
        container, name, profile
    )
    if existing_object is None:
        return state_result(True, "Object already absent", name, {})
    else:
        result = __salt__["libcloud_storage.delete_object"](container, name, profile)
        return state_result(result, "Deleted object", name, {})


def file_present(container, name, path, profile, overwrite_existing=False):
    """
    Ensures a object is downloaded locally.

    :param container: Container name
    :type  container: ``str``

    :param name: Object name in cloud
    :type  name: ``str``

    :param path: Local path to file
    :type  path: ``str``

    :param profile: The profile key
    :type  profile: ``str``

    :param overwrite_existing: Replace if already exists
    :type  overwrite_existing: ``bool``
    """
    result = __salt__["libcloud_storage.download_object"](
        path, container, name, profile, overwrite_existing
    )
    return state_result(result, "Downloaded object", name, {})
